# cqlstatement.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (CQL) statement parser.

Thes extracts from http://www.gadycosteff.com/cql/doc/operation.html for cql5.1
briefly describe CQL and how the cql.exe program operates.

A CQL statement has three parts:

    The word cql
    The cql parameters: a sequence of cql parameters enclosed in parentheses
    The cql body: a sequence of filters

CQL successively reads each game in its input PGN file. For each such game, CQL
plays through each move in the game.

Every time CQL reaches a position, it sets the current position to that
position. Then CQL tries to match each of its filters against that position.
If all the filters in the CQL file match the position, then CQL is said to have
matched the position.

When CQL has finished trying to match all the positions in the game against its
filters, then if any of the positions matched, CQL will output the game to the
output file. The output file can be specified either by default (in which case
its name is the name of the CQL file with -out appended); or as a parameter in
to the cql arguments in the CQL file itself; or on the command line.


The earlier partial position scheme implemented a sequence of piece designator
filters in CQL terms.  The equivalent of piece designator was limited to the
form 'Ra3', with pieces identified as one of 'KQRBNPkqrbnp?Xx-' where X is
replaced by A, x by a, ? by [Aa], and - by [Aa.], in CQL.  'A' is any white
piece, 'a' is any black piece, and '.' is an empty square. '[Aa.]' means it
does not matter whether the square is occupied because either white piece or
black piece or empty square matches.

"""

from chessql.core.statement import Statement, ErrorInformation

# from chessql.core.constants import PIECE_DESIGNATOR_FILTER
# from chessql.core.piecedesignator import PieceDesignator

from .cqlnode import CQLNode, FSNode
from .constants import NAME_DELIMITER

_ROTATE45_VALIDATION_TABLE = str.maketrans("12345678", "xxxxxxxx")


class CQLStatementError(Exception):
    """Exception class for cqlstatement module."""


class CQLStatement(Statement):
    """CQL statement parser.

    Parse text for a CQL statement.

    """

    create_node = CQLNode

    def __init__(self):
        """Delegate then initialize description and database name."""
        # Error handling is slightly different to .querystatement module.
        # The error information object is held directly here while it is held
        # in the Where instance used to calculate the query there.
        # This module processes instructions used to generate Where instances.
        super().__init__()
        self._description_string = ""

        # Not sure this is needed or wanted.
        # See datasource argument to _get_games_matching_filters().
        # dbset not used until a QueryStatement instance evaluates query.
        # self._dbset = None

        # _dbset and __database are passed to QueryStatement instances which
        # evaluate the query: not used directly in this class.

        self._dbset = None

        # Support using where.Where or where_dpt.Where depending on database
        # engine being used.
        # This attribute should not be used for anything else.
        self.__database = None
        # self.root_filter_node = None

    @property
    def dbset(self):
        """Return database filename."""
        return self._dbset

    @dbset.setter
    def dbset(self, value):
        if self._dbset is None:
            self._dbset = value
        elif self._dbset != value:
            raise CQLStatementError(
                "".join(
                    (
                        "Database file name already set to ",
                        repr(self._dbset),
                        ", cannot change to ",
                        repr(value),
                        ".",
                    )
                )
            )

    def set_database(self, database=None):
        """Set Database instance to which ChessQL query is applied."""
        self.__database = database

    def get_name_text(self):
        """Return name text."""
        return self._description_string

    def get_name_statement_text(self):
        """Return name and statement text."""
        return NAME_DELIMITER.join(
            (
                self._description_string,
                self.get_statement_text(),
            )
        )

    def process_statement(self, text):
        """Lex and parse the ChessQL statement.

        Two attempts to process the text are done.  The first treats the whole
        text as a ChessQL statement.  If it fails the second attempt treats
        the first line of text as the query's name and the rest as a ChessQL
        statement.

        Thus the query's name cannot be a valid CQL version 6 filter or
        sequence of filters.

        """
        self._description_string = ""
        super().process_statement(text.strip())
        if self.cql_error:
            rule = [t.strip() for t in text.split(NAME_DELIMITER, 1)]
            if len(rule) == 1:

                # The text cannot have an initial name component because text
                # equals rule[0] and there is no point in processing the text
                # again without the non-existent name component.
                # Assume initialization needed if text == ''.
                if text == "":
                    self._statement_string = ""
                    self._reset_state()
                    self._error_information = False
                return

            self._description_string = rule[0]
            super().process_statement(rule[1])

        # As it was in chessql.core.statement module, where it did not reject
        # any statements (possibly because rotate45 was never tried seriously)
        # before evaluation code transferred to ChessTab.
        # return self._rotate45_specific_squares(self.cql_filters, [])

        # self.cql_parameters and self.cql_filters contain a node tree for a
        # valid Chess QL statement.
        if self.cql_filters is None:
            return
        transformation = self.cql_filters.transform_piece_designators(
            FSNode(self.cql_filters)
        )
        if transformation:
            self.cql_error = transformation
        return

    def _rotate45_specific_squares(self, filter_, rotate45stack):
        if filter_.name == "rotate45":
            rotate45stack.append(None)
        for node in filter_.children:
            if rotate45stack and node.name == "piecedesignator":
                # if {c for c in node.leaf}.intersection('12345678'))
                if node.leaf != node.leaf.translate(
                    _ROTATE45_VALIDATION_TABLE
                ):
                    self._error_information = ErrorInformation(
                        self._statement_string.strip()
                    )
                    self._error_information.description = (
                        "rotate45 on specific squares"
                    )
                    return self._error_information
            r45ss = self._rotate45_specific_squares(node, rotate45stack)
            if r45ss:
                return r45ss
        if filter_.name == "rotate45":
            rotate45stack.pop()
        return None



# chessdbtkinter.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database using Berkeley DB.

The Berkeley DB interface has significantly worse performance than DPT when
doing multi-index searches.  However it is retained since DPT became available
because tracking down problems in the chess logic using IDLE can be easier
in the *nix environment.
"""

import os

from solentware_base import db_tkinter_database

from ..core.filespec import (
    FileSpec,
    DB_ENVIRONMENT_GIGABYTES,
    DB_ENVIRONMENT_BYTES,
    DB_ENVIRONMENT_MAXLOCKS,
)
from ..basecore import database


class ChessDatabase(database.Database, db_tkinter_database.Database):
    """Provide access to a database of games of chess."""

    _deferred_update_process = os.path.join(
        os.path.basename(os.path.dirname(__file__)), "runchessdbtkinterdu.py"
    )

    def __init__(
        self,
        DBfile,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        Arguments are passed through to superclass __init__.

        """
        dbnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )

        environment = {
            "flags": (
                "-create",
                "-recover",
                "-txn",
                "-private",
                "-system_mem",
            ),
            "gbytes": DB_ENVIRONMENT_GIGABYTES,
            "bytes": DB_ENVIRONMENT_BYTES,
            "maxlocks": DB_ENVIRONMENT_MAXLOCKS,
        }

        super().__init__(
            dbnames, folder=DBfile, environment=environment, **kargs
        )

    def _delete_database_names(self):
        """Override and return tuple of filenames to delete."""
        return (self.database_file, self._get_log_dir_name())



# constants.py
# Copyright 2010 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Constants for Chess Query Language (CQL) parser.

Uses pgn_read.core.constants values plus additional piece encodings for any
piece, any white piece, and so on.

These constants were used by the old partial position scheme, which implemented
a single position list in CQL terms.

"""

from pgn_read.core.constants import (
    FEN_WHITE_PAWN,
    FEN_BLACK_PAWN,
    TAG_RESULT,
    DEFAULT_TAG_RESULT_VALUE,
    DEFAULT_TAG_VALUE,
    TAG_FEN,
    TAG_SETUP,
    SETUP_VALUE_FEN_PRESENT,
)

from chessql.core.constants import (
    ANY_WHITE_PIECE_NAME,
    ANY_BLACK_PIECE_NAME,
    EMPTY_SQUARE_NAME,
    WHITE_PIECE_NAMES,
    BLACK_PIECE_NAMES,
)

# Composite piece map (CQL) to actual pieces (PGN).
MAP_CQL_PIECE_TO_PIECES = {
    ANY_WHITE_PIECE_NAME: WHITE_PIECE_NAMES,
    ANY_BLACK_PIECE_NAME: BLACK_PIECE_NAMES,
    EMPTY_SQUARE_NAME: WHITE_PIECE_NAMES + BLACK_PIECE_NAMES,
}

NAME_DELIMITER = "\n"
BOARDSIDE = 8
BOARDSQUARES = BOARDSIDE * BOARDSIDE
PIECE_SQUARE_NOT_ALLOWED = set()
for _piece in FEN_WHITE_PAWN, FEN_BLACK_PAWN:
    for _square in range(BOARDSIDE):
        PIECE_SQUARE_NOT_ALLOWED.add((_piece, _square))
        PIECE_SQUARE_NOT_ALLOWED.add((_piece, BOARDSQUARES - _square - 1))
PIECE_SQUARE_NOT_ALLOWED = frozenset(PIECE_SQUARE_NOT_ALLOWED)

# PGN constants for repertoires.
TAG_OPENING = "Opening"
REPERTOIRE_TAG_ORDER = (TAG_OPENING, TAG_RESULT)
REPERTOIRE_GAME_TAGS = {
    TAG_OPENING: DEFAULT_TAG_VALUE,
    TAG_RESULT: DEFAULT_TAG_RESULT_VALUE,
}

# Lookup for move number component of key values in game indicies: covers the
# likely values for typical game PGN without recursive annotation variations.
MOVE_NUMBER_KEYS = tuple(
    ["0"] + [str(len(hex(i)) - 2) + hex(i)[2:] for i in range(1, 256)]
)

# Character representation of empty square on displayed board.
NOPIECE = ""

# PGN results.  pgn_read.core.constants has '*' as DEFAULT_TAG_RESULT_VALUE
# but does not have constants for the other three results.
WHITE_WIN = "1-0"
BLACK_WIN = "0-1"
DRAW = "1/2-1/2"
UNKNOWN_RESULT = "*"

# Start and end tag characters.
END_TAG = "]"
START_TAG = "["

# Start and end RAV characters.
END_RAV = ")"
START_RAV = "("

# Start and end comment characters.
END_COMMENT = "}"
START_COMMENT = "{"
START_EOL_COMMENT = ";"

# Decorators to do special cases for Date and Round sorting.
SPECIAL_TAG_DATE = ("?", "0")

# Variation markers and non-move placeholders.
NON_MOVE = None

# Error markers for PGN display.
ERROR_START_COMMENT = "Error::"
ESCAPE_END_COMMENT = "::" + START_COMMENT + START_COMMENT + "::"
HIDE_END_COMMENT = "::::" + START_COMMENT + START_COMMENT

FEN_CONTEXT = (
    "".join((START_TAG, TAG_FEN, '"')),
    "".join(
        (
            '"',
            END_TAG,
            START_TAG,
            TAG_SETUP,
            '"',
            SETUP_VALUE_FEN_PRESENT,
            END_TAG.join('"\n'),
        )
    ),
)

del _piece, _square
del FEN_WHITE_PAWN, FEN_BLACK_PAWN, BOARDSQUARES
del ANY_WHITE_PIECE_NAME, ANY_BLACK_PIECE_NAME, EMPTY_SQUARE_NAME
del WHITE_PIECE_NAMES, BLACK_PIECE_NAMES
del TAG_FEN, TAG_SETUP, SETUP_VALUE_FEN_PRESENT



# chessboard.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Demonstrate chess board class and methods to draw position on board."""


if __name__ == "__main__":

    import tkinter

    from pgn_read.core.piece import Piece
    from pgn_read.core.constants import (
        FEN_WHITE_KING,
        FEN_WHITE_QUEEN,
        FEN_WHITE_ROOK,
        FEN_WHITE_BISHOP,
        FEN_WHITE_KNIGHT,
        FEN_WHITE_PAWN,
        FEN_BLACK_KING,
        FEN_BLACK_QUEEN,
        FEN_BLACK_ROOK,
        FEN_BLACK_BISHOP,
        FEN_BLACK_KNIGHT,
        FEN_BLACK_PAWN,
    )

    from ..gui import fonts
    from ..gui.board import Board
    from ..core.constants import NOPIECE

    root = tkinter.Tk()
    root.wm_title("Demonstrate Board")
    f = fonts.make_chess_fonts(root, preferred_pieces=("Chess Lucena",))
    b = Board(root, boardborder=10)
    del f
    b.get_top_widget().pack(fill=tkinter.BOTH, expand=tkinter.TRUE)
    b.get_top_widget().pack_propagate(False)
    b.set_board(
        {
            "a8": Piece(FEN_BLACK_ROOK, "a8"),
            "b8": Piece(FEN_BLACK_KNIGHT, "b8"),
            "c8": Piece(FEN_BLACK_BISHOP, "c8"),
            "d8": Piece(FEN_BLACK_QUEEN, "d8"),
            "e8": Piece(FEN_BLACK_KING, "e8"),
            "f8": Piece(FEN_BLACK_BISHOP, "f8"),
            "g8": Piece(FEN_BLACK_KNIGHT, "g8"),
            "h8": Piece(FEN_BLACK_ROOK, "h8"),
            "a7": Piece(FEN_BLACK_PAWN, "a7"),
            "b7": Piece(FEN_BLACK_PAWN, "b7"),
            "c7": Piece(FEN_BLACK_PAWN, "c7"),
            "d7": Piece(FEN_BLACK_PAWN, "d7"),
            "e7": Piece(FEN_BLACK_PAWN, "e7"),
            "f7": Piece(FEN_BLACK_PAWN, "f7"),
            "g7": Piece(FEN_BLACK_PAWN, "g7"),
            "h7": Piece(FEN_BLACK_PAWN, "h7"),
            "a2": Piece(FEN_WHITE_PAWN, "a2"),
            "b2": Piece(FEN_WHITE_PAWN, "b2"),
            "c2": Piece(FEN_WHITE_PAWN, "c2"),
            "d2": Piece(FEN_WHITE_PAWN, "d2"),
            "e2": Piece(FEN_WHITE_PAWN, "e2"),
            "f2": Piece(FEN_WHITE_PAWN, "f2"),
            "g2": Piece(FEN_WHITE_PAWN, "g2"),
            "h2": Piece(FEN_WHITE_PAWN, "h2"),
            "a1": Piece(FEN_WHITE_ROOK, "a1"),
            "b1": Piece(FEN_WHITE_KNIGHT, "b1"),
            "c1": Piece(FEN_WHITE_BISHOP, "c1"),
            "d1": Piece(FEN_WHITE_QUEEN, "d1"),
            "e1": Piece(FEN_WHITE_KING, "e1"),
            "f1": Piece(FEN_WHITE_BISHOP, "f1"),
            "g1": Piece(FEN_WHITE_KNIGHT, "g1"),
            "h1": Piece(FEN_WHITE_ROOK, "h1"),
        }
    )
    del b
    root.pack_propagate(False)
    root.mainloop()



# rayfilter.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (ChessQL) ray filter evaluator.

Examples of ray filters are 'ray ( Q n k )' and 'ray ( Q[a4,c3] n kh1-8 )'.

RayFilter expands the list of square descriptions into the particular rays,
horizontal, vertical, and diagonal, which need to be evaluated.

The CQL definition of ray filter allows any set filter, not just piece
designators, in the ray filter.  For example ray ( Q up 1 n k).

The RayFilter class ignores, at best, other set filters: in the example
'up 1 n' is treated as 'n' ignoring the 'up' filter.

"""
from chessql.core import constants
from chessql.core import piecedesignator
from chessql.core.cql import Token
from chessql.core import rays

from ..core.constants import MOVE_NUMBER_KEYS


class RayFilterError(Exception):
    """Exception class for rayfilter module."""


class RayFilter:
    """ChessQL ray filter evaluator.

    The ray filter has a list of square specifiers, piece designators
    usually, which define the rays to be evaluated.

    This class assumes the caller has expanded the piece designator parameters
    to the ray or between filter; and applied any transforms.

    Use this class like:
    C(RayFilter)

    Subclasses must implement the database interface specific methods defined
    in this class which raise RayFilterError('Not implemented')
    exceptions.

    """

    def __init__(self, filter_):
        """Initialize to apply filter_ on move_number."""
        if filter_.tokendef is not Token.RAY:
            raise RayFilterError(
                "".join(
                    (
                        "Filter '",
                        filter_.name,
                        "' is not a ray filter.",
                    )
                )
            )

        if len(filter_.children) < 2:
            raise RayFilterError(
                "".join(
                    (
                        "Filter '",
                        filter_.name,
                        "' must have at least two arguments.",
                    )
                )
            )

        raysets = []
        for node in filter_.children:
            designator_set = set()
            raysets.append(designator_set)
            stack = [node]
            while stack:
                if stack[-1].tokendef is Token.PIECE_DESIGNATOR:
                    designator_set.update(
                        piece_square_to_index(stack[-1].data.designator_set)
                    )
                    stack.pop()
                    continue
                topnode = stack.pop()
                for spc in topnode.children:
                    stack.append(spc)
        raysets = tuple(tuple(i) for i in raysets)
        raycomponents = []
        size = tuple(len(i) for i in raysets)
        count = 1
        for elementsize in size:
            count *= elementsize
        looper = [i - 1 for i in size]
        while count:
            count -= 1
            self._append_linear_rayitem_to_raycomponents(
                [raysets[e][i] for e, i in enumerate(looper)], raycomponents
            )
            for element, item in enumerate(looper):
                if not item:
                    continue
                looper[element] -= 1
                if element:
                    for looperitem in range(element):
                        looper[looperitem] = size[looperitem] - 1
                break
        self.raycomponents = raycomponents

    @staticmethod
    def _append_linear_rayitem_to_raycomponents(rayitem, raycomponents):
        """Add linear rayitem to raycomponents list for evaluation.

        The empty square specifications, if any, are added.

        """
        start_square = rayitem[0][:-1]
        final_square = rayitem[-1][:-1]
        ray_squares = rays.get_ray(start_square, final_square)
        if ray_squares is None:
            return
        rayitem_set = set(r[:-1] for r in rayitem[1:-1])
        ray_squares_set = set(ray_squares[1:-1])
        if rayitem_set.difference(ray_squares_set):
            return
        for square in ray_squares_set.difference(rayitem_set):
            rayitem.append(square + constants.EMPTY_SQUARE_NAME)
        if start_square > final_square:
            raycomponents.append(list(reversed(sorted(rayitem))))
        else:
            raycomponents.append(sorted(rayitem))

    def find_games(self, movenumber, variation, evaluator):
        """Find games fitting ray for movenumber and variation."""
        recordset_cache = {}
        recordset_count_cache = {}
        prefix = move_number_str(movenumber) + variation
        for item in self.raycomponents:
            raycomponent = [prefix + i for i in item]
            for mvsp in raycomponent:
                if mvsp not in recordset_cache:
                    recordset_cache[mvsp] = evaluator(mvsp)
                    recordset_count_cache[mvsp] = recordset_cache[
                        mvsp
                    ].count_records()
        for count in recordset_count_cache.values():
            if not count:
                return {}
        return recordset_cache


# This function belong in, and has been moved to, a chesstab module.  It came
# from chessql.core.piecedesignator.PieceDesignator, it was a staticmethod, but
# chesstab has no subclass of of PieceDesignator (yet), and rayfilter is only
# user at present.
# The function may go back to chessql now that the movenumber and variation
# elements have been removed: these are chesstab implementation details and
# what remains is properly chessql stuff.
# The alternative commented definition with the <piece square> order has been
# removed because the <square piece> order is more convenient for sorting
# ray components into square order too.
# The designator set names <piece squares> because that is how they are
# expressed in CQL statements.
def piece_square_to_index(designator_set):
    """Convert piece designator set values to index format: Qa4 to a4Q.

    Assumed that having all index values for a square adjacent is better
    than having index values for piece together, despite the need for
    conversion.

    """
    file_names = constants.FILE_NAMES
    rank_names = constants.CQL_RANK_NAMES
    ecs = piecedesignator.PieceDesignator.expand_composite_square
    indexset = set()
    for piece_square in designator_set:
        if len(piece_square) != 1:
            indexset.add(piece_square[1:] + piece_square[0])
        else:
            indexset.update(
                {
                    s + piece_square
                    for s in ecs(
                        file_names[0],
                        file_names[-1],
                        rank_names[0],
                        rank_names[-1],
                    )
                }
            )
    return indexset


def move_number_str(move_number):
    """Return hex(move_number) values prefixed with string length.

    A '0x' prefix is removed first.

    """
    # Adapted from module pgn_read.core.parser method add_move_to_game().
    try:
        return MOVE_NUMBER_KEYS[move_number]
    except IndexError:
        base16 = hex(move_number)
        return str(len(base16) - 2) + base16[2:]



# chessdu.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Define User Interface for deferred update process."""

import sys
import os
import traceback
import datetime
import tkinter
import tkinter.font
import tkinter.messagebox
import queue
import time

from solentware_misc.core import callthreadqueue
from solentware_misc.gui.tasklog import LogText

from solentware_bind.gui.bindings import Bindings

from pgn_read.core.parser import PGN

from ..core.pgn import GameUpdateEstimate
from .. import (
    ERROR_LOG,
    APPLICATION_NAME,
)
from ..core.filespec import GAMES_FILE_DEF

# Time taken to parse a sample of a PGN file is measured.
# Number of games in file is estimated from number of bytes used in game scores
# compared with number of bytes in file.
# _DATABASE_UPDATE_FACTOR is set from measuring time taken to import a PGN file
# containing a large number of games to an empty database.  Large means over a
# million games.
# This factor is a big under-estimate when number of games is less than segment
# size.  In practice the difference will be noticed when importing less than
# 131072 games to a database which already has games: the new games will span
# up to three segments when imported.
# Available memory has a big impact because it determines how long the runs of
# sequential updates to indexes can be.  The factor 5 is appropriate when at
# least 1.5Gb is available.  A default build of FreeBSD 10.1 on a PC with 2Gb
# installed passes this test; but Microsoft Windows XP, and later presumably,
# needs more memory to do so.  A default build of OpenBSD 5.9 restricts user
# processes to 0.5Gb.  The situation is not known for OS X or any Linux
# distribution.
# Factor changed from 3 to 5 when CQL5.1 syntax introduced to implement partial
# position searches, due to extra index updates.
_DATABASE_UPDATE_FACTOR = 5


class ChessDeferredUpdate(Bindings):
    """Connect a chess database with User Interface for deferred update."""

    def __init__(
        self, deferred_update_method=None, database_class=None, sample=5000
    ):
        """Create the database and ChessUI objects.

        deferred_update_method - the method to do the import
        database_class - access the database with an instance of this class
        sample - estimate size of import from first 'sample' games in PGN file

        """
        super().__init__()
        self.queue = callthreadqueue.CallThreadQueue()
        self.reportqueue = queue.Queue(maxsize=1)

        self.dumethod = deferred_update_method
        self.sample = sample
        self.estimate_data = None
        self._allow_job = None
        self._import_job = None
        self.database = database_class(
            sys.argv[1], allowcreate=True, deferupdatefiles={GAMES_FILE_DEF}
        )

        self.root = tkinter.Tk()
        self.root.wm_title(
            " - ".join(
                (
                    " ".join((APPLICATION_NAME, "Import")),
                    os.path.basename(sys.argv[1]),
                )
            )
        )
        frame = tkinter.Frame(master=self.root)
        frame.pack(side=tkinter.BOTTOM)
        self.buttonframe = tkinter.Frame(master=frame)
        self.buttonframe.pack(side=tkinter.BOTTOM)

        # See comment near end of this class definition for explanation of this
        # change.  It was simpler to display all the buttons at once and add
        # some control logic, than wrap the things in a task queue for the main
        # thread to process.  But the presence or absence of the buttons should
        # be the control logic.
        self._quit_message = "".join(
            (
                "The import has not been started.",
                "\n\nDo you want to abandon the import?",
            )
        )
        self.cancel = tkinter.Button(
            master=self.buttonframe,
            # text='Cancel',
            text="Quit",
            underline=0,
            command=self.try_command(
                # self.quit_before_import_started, self.buttonframe))
                self._quit_import,
                self.buttonframe,
            ),
        )
        self.cancel.pack(side=tkinter.RIGHT, padx=12)
        backup = tkinter.Button(
            master=self.buttonframe,
            text="Import with Backups",
            underline=12,
            command=self.try_command(
                self._do_import_with_backup, self.buttonframe
            ),
        )
        backup.pack(side=tkinter.RIGHT, padx=12)
        import_ = tkinter.Button(
            master=self.buttonframe,
            text="Import",
            underline=0,
            command=self.try_command(
                self._do_import_without_backup, self.buttonframe
            ),
        )
        import_.pack(side=tkinter.RIGHT, padx=12)

        # See comment near end of this class definition for explanation of this
        # change.  Replacing Text by LogText fixes the majority of the
        # problems.
        # Consequential changes, self.append_text to self.report.append_text
        # mainly, are not marked.
        # Added 07 August 2016:
        # Edited 12 October 2022 when all tkinter.Text arguments put in cnf.
        # _get_default_font_actual() value used as cnf to avoid picking fonts
        # like 'Chess Cases' as default font, when running under Wine, from
        # ~/.wine/drive_c/windows/Fonts or /usr/local/share/fonts.
        # The _get_default_font_actual() value is not needed in cnf otherwise.
        # cnf = _get_default_font_actual(tkinter.Text) is not accepted as an
        # argument by tkinter.Text().  It seems this argument was ignored
        # previously en-route to the Text() call.
        self.report = LogText(
            master=self.root,
            get_app=self,
            cnf=dict(wrap=tkinter.WORD, undo=tkinter.FALSE),
        )
        self.report.focus_set()
        self.bind(
            self.report,
            "<Alt-b>",
            function=self.try_event(self._do_import_with_backup),
        )
        self.bind(
            self.report,
            "<Alt-i>",
            function=self.try_event(self._do_import_without_backup),
        )
        self.bind(
            self.report, "<Alt-q>", function=self.try_event(self._quit_import)
        )
        self.database.add_import_buttons(
            self.buttonframe,
            self.try_command,
            self.try_event,
            self.bind,
            self.report,
        )

        self.report.tag_configure(
            "margin",
            lmargin2=tkinter.font.nametofont(self.report.cget("font")).measure(
                "2010-05-23 10:20:57  "
            ),
        )
        self.tagstart = "1.0"
        self.report.append_text(
            "".join(("Importing to database ", sys.argv[1], "."))
        )
        self.report.append_text(
            "All times quoted assume no other applications running.",
            timestamp=False,
        )
        self.report.append_text_only("")
        self.report.append_text("Estimating number of games in import.")
        self.report.append_text_only("")
        self.report.pack(
            side=tkinter.LEFT, fill=tkinter.BOTH, expand=tkinter.TRUE
        )
        self.root.iconify()
        self.root.update()
        self.root.deiconify()
        self.__run_ui_task_from_queue(1000)
        self.root.after(100, self.try_command(self._run_allow, self.root))
        self.root.mainloop()

    def _allow_import(self):
        """Do checks for database engine and return True if import allowed."""
        # The close_database() in finally clause used to be the first statement
        # after runjob() definition in _run_import() method.  An exception was
        # raised using the sqlite3 module because run_input() is run in a
        # different thread from allow_input().  Earlier versions of chessdu did
        # not attempt to close the connection, hiding the problem.
        # The apsw module did not raise an exception, nor did modules providing
        # an interface to Berkeley DB or DPT.
        self.database.open_database()
        try:
            if not self._estimate_games_in_import():
                self.report.append_text("There are no games to import.")
                self.report.append_text_only("")
                return None
            if self._allow_time():
                self.database.report_plans_for_estimate(
                    self._get_pgn_file_estimates(), self.report
                )
                return True
            self.report.append_text("Unable to estimate time to do import.")
            self.report.append_text_only("")
            return False
        finally:
            self.database.close_database()

    def _allow_time(self):
        """Ask is deferred update to proceed if game count is estimated.

        The time taken will vary significantly depending on environment.

        """
        if not self.estimate_data:
            return False
        seconds = (
            (self.estimate_data[0] + self.estimate_data[4])
            * self.estimate_data[9]
            * _DATABASE_UPDATE_FACTOR
        )
        minutes, seconds = divmod(round(seconds), 60)
        hours, minutes = divmod(minutes, 60)
        days, hours = divmod(hours, 24)
        duration = []
        if days:
            duration.append(str(days))
            duration.append("days")
            if hours:
                duration.append(str(hours))
                duration.append("hours")
        elif hours:
            duration.append(str(hours))
            duration.append("hours")
            if minutes:
                duration.append(str(minutes))
                duration.append("minutes")
        elif minutes:
            duration.append(str(minutes))
            duration.append("minutes")
            if seconds:
                duration.append(str(seconds))
                duration.append("seconds")
        elif seconds > 1:
            duration.append(str(seconds))
            duration.append("seconds")
        else:
            duration.append(str(1))
            duration.append("second")
        self.report.append_text(
            "".join(
                (
                    "The estimate is the time taken to process the sample ",
                    "scaled up to the estimated number of games then ",
                    "multiplied by ",
                    str(_DATABASE_UPDATE_FACTOR),
                    ".  Expect the import to take longer if the database ",
                    "already contains games: the effect is worse for smaller "
                    "imports.  Progress reports are made if the import is ",
                    "large enough.",
                )
            )
        )
        self.report.append_text_only("")
        self.report.append_text(
            "".join(
                ("The import is expected to take ", " ".join(duration), ".")
            )
        )
        self.report.append_text_only("")
        return True

    def _do_import_with_backup(self, event=None):
        """Run import thread if allowed and not already run.

        event is ignored and is present for compatibility between button click
        and keypress.

        """
        del event
        if self._import_job:
            return
        names, exists = self.database.get_archive_names(
            files=(GAMES_FILE_DEF,)
        )
        if exists:
            if not tkinter.messagebox.askokcancel(
                parent=self.root,
                title="Import Backup",
                message="".join(
                    (
                        "Import backups of the following files already ",
                        "exist.\n\n",
                        "\n".join(exists),
                        "\n\nThis may mean an earlier Import action failed ",
                        'and has not yet been resolved.\n\nClick "Ok" to ',
                        "delete these backups and create new ones before ",
                        "doing the Import.  These backups will be deleted ",
                        "if the Import succeeds.\n\nClick ",
                        '"Cancel" to leave things as they are.',
                    )
                ),
            ):
                return
            self.report.append_text(
                "Backups will be taken before doing the import."
            )
            self.report.append_text(
                "These backups will replace the existing backups.",
                timestamp=False,
            )
        elif not tkinter.messagebox.askokcancel(
            parent=self.root,
            title="Import Backup",
            message="".join(
                ("Please confirm the import is to be done with backups.",)
            ),
        ):
            return
        else:
            self.report.append_text(
                "Backups will be taken before doing the import."
            )
        self.report.append_text(
            "These backups will be deleted if the import succeeds.",
            timestamp=False,
        )
        self.report.append_text_only("")
        self._run_import(backup=True, names=names)

    def _do_import_without_backup(self, event=None):
        """Run import thread if allowed and not already run.

        event is ignored and is present for compatibility between button click
        and keypress.

        """
        del event
        if self._import_job:
            return
        names, exists = self.database.get_archive_names(
            files=(GAMES_FILE_DEF,)
        )
        if exists:
            if not tkinter.messagebox.askokcancel(
                parent=self.root,
                title="Import Backup",
                message="".join(
                    (
                        "Import backups of the following files already ",
                        "exist.\n\n",
                        "\n".join(exists),
                        "\n\nThis may mean an earlier Import action failed ",
                        'and has not yet been resolved.\n\nClick "Ok" to ',
                        "delete these backups and create new ones before ",
                        "doing the Import.  These backups will be deleted ",
                        "if the Import succeeds.\n\nClick ",
                        '"Cancel" to leave things as they are.',
                    )
                ),
            ):
                return
            self.report.append_text(
                "The import will be done without taking backups."
            )
            self.report.append_text_only("")
            self.report.append_text(
                "Existing backups will be deleted before doing the import.",
                timestamp=False,
            )
            self.report.append_text_only("")
        elif not tkinter.messagebox.askokcancel(
            parent=self.root,
            title="Import Backup",
            message="".join(
                ("Please confirm the import is to be done without backups.",)
            ),
        ):
            # self.report.append_text(
            #     "The import will be done without taking backups."
            # )
            # self.report.append_text_only("")
            return
        self._run_import(backup=False, names=names)

    def _estimate_games_in_import(self):
        """Estimate import size from the first sample games in import files."""
        self.estimate_data = False
        text_file_size = sum([os.path.getsize(pp) for pp in sys.argv[2:]])
        reader = PGN(game_class=GameUpdateEstimate)
        errorcount = 0
        totallen = 0
        totalerrorlen = 0
        totalgamelen = 0
        gamecount = 0
        positioncount = 0
        piecesquaremovecount = 0
        piecemovecount = 0
        estimate = False
        time_start = time.monotonic()
        for pgnfile in sys.argv[2:]:
            if gamecount + errorcount >= self.sample:
                estimate = True
                break
            with open(pgnfile, "r", encoding="iso-8859-1") as source:
                for rcg in reader.read_games(source):
                    if gamecount + errorcount >= self.sample:
                        estimate = True
                        break
                    if len(rcg.pgn_text):
                        rawtokenlen = rcg.end_char - rcg.start_char
                    else:
                        rawtokenlen = 0
                    if rcg.state is not None:
                        errorcount += 1
                        totalerrorlen += rawtokenlen
                    else:
                        gamecount += 1
                        totalgamelen += rawtokenlen
                        positioncount += len(rcg.positionkeys)
                        piecesquaremovecount += len(rcg.piecesquaremovekeys)
                        piecemovecount += len(rcg.piecemovekeys)
                    totallen += rawtokenlen
        time_end = time.monotonic()
        if estimate:
            try:
                scale = float(text_file_size) // totallen
            except ZeroDivisionError:
                scale = 0
        else:
            scale = 1
        try:
            bytes_per_game = totalgamelen // gamecount
        except ZeroDivisionError:
            bytes_per_game = 0
        try:
            bytes_per_error = totalerrorlen // errorcount
        except ZeroDivisionError:
            bytes_per_error = 0
        try:
            positions_per_game = positioncount // gamecount
        except ZeroDivisionError:
            positions_per_game = 0
        try:
            pieces_per_game = piecesquaremovecount // gamecount
        except ZeroDivisionError:
            pieces_per_game = 0
        try:
            piecetypes_per_game = piecemovecount // gamecount
        except ZeroDivisionError:
            piecetypes_per_game = 0

        self.estimate_data = (
            int(gamecount * scale),
            bytes_per_game,
            positions_per_game,
            pieces_per_game,
            int(errorcount * scale),
            bytes_per_error,
            estimate,
            gamecount,
            errorcount,
            (time_end - time_start) / (gamecount + errorcount),
        )
        if estimate:
            self.report.append_text(
                " ".join(("Estimated Games:", str(self.estimate_data[0])))
            )
            self.report.append_text(
                " ".join(("Estimated Errors:", str(self.estimate_data[4]))),
                timestamp=False,
            )
            self.report.append_text(
                " ".join(("Sample Games:", str(gamecount))), timestamp=False
            )
            self.report.append_text(
                " ".join(("Sample Errors:", str(errorcount))), timestamp=False
            )
        else:
            self.report.append_text(
                "The import is small so all games are counted."
            )
            self.report.append_text(
                " ".join(("Games in import:", str(gamecount))), timestamp=False
            )
            self.report.append_text(
                " ".join(("Errors in import:", str(errorcount))),
                timestamp=False,
            )
        self.report.append_text(
            " ".join(("Bytes per game:", str(bytes_per_game))), timestamp=False
        )
        self.report.append_text(
            " ".join(("Bytes per error:", str(bytes_per_error))),
            timestamp=False,
        )
        self.report.append_text(
            " ".join(("Positions per game:", str(positions_per_game))),
            timestamp=False,
        )

        # positions_per_game == 0 if there are no valid games.
        ppp = "Not a number"
        for count, text in (
            (pieces_per_game, "Pieces per position:"),
            (piecetypes_per_game, "Piece types per position:"),
        ):
            if positions_per_game:
                ppp = str(count // positions_per_game)
            self.report.append_text(" ".join((text, ppp)), timestamp=False)
        self.report.append_text("", timestamp=False)

        # Check if import can proceed
        if gamecount + errorcount == 0:
            self.report.append_text(
                "No games, or games with errors, found in import."
            )
            return False
        if errorcount == 0:
            if estimate:
                self.report.append_text("No games with errors in sample.")
                self.report.append_text(
                    " ".join(
                        (
                            "It is estimated no games with",
                            "errors exist in import.",
                        )
                    ),
                    timestamp=False,
                )
                self.report.append_text(
                    "Any found in import will be indexed only as errors.",
                    timestamp=False,
                )
            else:
                self.report.append_text("No games with errors in import.")
            self.report.append_text_only("")
        elif estimate:
            self.report.append_text(
                "Games with errors have been found in sample."
            )
            self.report.append_text(
                " ".join(
                    (
                        "The sample is the first",
                        str(self.sample),
                        "games in import.",
                    )
                ),
                timestamp=False,
            )
            self.report.append_text(
                "All found in import will be indexed only as errors.",
                timestamp=False,
            )
            self.report.append_text_only("")
        else:
            self.report.append_text(
                "Games with errors have been found in sample."
            )
            self.report.append_text(
                "All found in import will be indexed only as errors.",
                timestamp=False,
            )
            self.report.append_text_only("")
        return bool(self.estimate_data)

    # Override ChessException method as ChessUI class is not used.
    # May be wrong now solentware_misc Bindings is used.
    def get_error_file_name(self):
        """Return the exception report file name."""
        return os.path.join(sys.argv[1], ERROR_LOG)

    def _get_pgn_file_estimates(self):
        """Return the estimates of object counts for a PGN file."""
        return self.estimate_data

    def quit_after_import_started(self):
        """Quit process after import has started."""
        if tkinter.messagebox.askyesno(
            parent=self.root,
            title="Abandon Import",
            message="".join(
                (
                    "The import has been started.\n\n",
                    APPLICATION_NAME,
                    " will attempt ",
                    "to restore the files using the backups if they were ",
                    "taken.\n\nDo you want to abandon the import?",
                )
            ),
        ):
            self.root.destroy()

    def _quit_import(self, event=None):
        """Quit process.

        event is ignored and is present for compatibility between button click
        and keypress.

        """
        del event
        if tkinter.messagebox.askyesno(
            parent=self.root, title="Quit Import", message=self._quit_message
        ):
            self.root.destroy()

    def _run_allow(self):
        """Run import thread if allowed and not already run."""
        if self._allow_job:
            return
        self._allow_job = True
        self.queue.put_method(self.try_thread(self._allow_import, self.root))

    def _run_import(self, backup=None, names=None):
        """Invoke method to do the deferred update and display job status.

        backup - the answer given to the 'take backups' prompt
        names - the files to backup if requested

        Closing the job status widget terminates the job.

        """
        if self._import_job:
            return

        def runjob(*a):
            try:
                if backup:
                    self.report.append_text("Taking backups.")
                    if self.database.archive(flag=backup, names=names) is None:
                        self.report.append_text("Taking backups abandonned")
                        self.report.append_text(
                            "".join(
                                (
                                    "The existing backup does not ",
                                    "look like a ",
                                    APPLICATION_NAME,
                                    " backup.",
                                )
                            ),
                            timestamp=False,
                        )
                        self.report.append_text_only("")
                        # self.cancel.configure(
                        #    command=self.try_command(
                        #        self.quit_before_import_started, self.cancel))
                        self._quit_message = "".join(
                            (
                                "The import has not been started.",
                                "\n\nDo you want to abandon the import?",
                            )
                        )
                        return
                    self.report.append_text("Backups saved.")
                    self.report.append_text_only("")
                self.report.append_text("Import started.")
                self.report.append_text_only("")
                status = False
                try:
                    status = self.dumethod(
                        *a, estimated_number_of_games=self.estimate_data[0]
                    )
                except Exception as error:
                    try:
                        write_error_to_log()
                    except Exception as error:
                        raise SystemExit(
                            " reporting exception in ".join(
                                ("Exception while", APPLICATION_NAME)
                            )
                        ) from error
                    raise SystemExit(
                        "Exception reported in " + APPLICATION_NAME
                    ) from error
                self.report.append_text("Import finished.")
                self.report.append_text_only("")
                if not status:
                    if backup:
                        self.report.append_text(
                            "".join(
                                (
                                    "Problem encountered doing import. ",
                                    "Backups are retained for recovery.",
                                )
                            )
                        )
                    else:
                        self.report.append_text(
                            "".join(
                                (
                                    "Problem encountered doing import. ",
                                    "No backups available.",
                                )
                            )
                        )
                    self.report.append_text_only(
                        "".join(("See ", ERROR_LOG, " for details."))
                    )
                elif backup:
                    self.report.append_text("Deleting backups.")
                    if (
                        self.database.delete_archive(flag=backup, names=names)
                        is None
                    ):
                        self.report.append_text("Deleting backups abandonned.")
                        self.report.append_text_only(
                            "".join(
                                (
                                    "The existing backup does not ",
                                    "look like a ",
                                    APPLICATION_NAME,
                                    " backup.",
                                )
                            )
                        )
                    else:
                        self.report.append_text("Backups deleted.")
                # self.cancel.configure(
                #    command=self.try_command(
                #        self.quit_after_import_finished, self.cancel),
                #    text='Quit')
                self._quit_message = "".join(
                    (
                        "The import has been completed.",
                        "\n\nDo you want to dismiss the import log?",
                    )
                )
            except Exception as error:
                try:
                    write_error_to_log()
                except Exception as error:
                    raise SystemExit(
                        " reporting exception in ".join(
                            ("Exception while", APPLICATION_NAME)
                        )
                    ) from error
                raise SystemExit(
                    "Exception reported in " + APPLICATION_NAME
                ) from error

        self._import_job = True
        self.queue.put_method(
            self.try_thread(runjob, self.root),
            args=(
                sys.argv[1],
                sys.argv[2:],
                self.database.get_file_sizes(),
                self.report.append_text,
            ),
        )

    # Methods __call__, get_reportqueue, __run_ui_task_from_queue, and
    # get_thread_queue, introduced to allow this module to work if tkinter is
    # compiled without --enable-threads as in OpenBSD 5.7 i386 packages.  The
    # standard build from FreeBSD ports until early 2015 at least, when this
    # change was introduced, is compiled with --enable-threads so the unchanged
    # code worked.  Not sure if the change in compiler on FreeBSD from gcc to
    # clang made a difference.  The Microsoft Windows' Pythons seem to be
    # compiled with --enable-threads because the unchanged code works in that
    # environment.  The situation on OS X, and any GNU-Linux distribution, is
    # not known.

    # Code in the solentware_misc.gui.tasklog module already dealt with this
    # problem, so the minimum necessary was copied to here.  The classes in
    # tasklog are modified versions of code present in this module before this
    # change.

    def __call__(self):
        """Return self allowing the construct <instance>()."""
        return self

    def get_reportqueue(self):
        """Return queue from which reports to be displayed are taken.

        Tasks run from self.queue place responses on this queue.

        """
        return self.reportqueue

    def get_thread_queue(self):
        """Return queue on which tasks to be run are placed.

        A task runner takes tasks from this queue.

        """
        return self.queue

    def __run_ui_task_from_queue(self, interval):
        """Do all queued tasks then wake-up after interval."""
        while True:
            try:
                method, args, kwargs = self.reportqueue.get_nowait()
                method(*args, **kwargs)
            except queue.Empty:
                self.root.after(
                    interval,
                    self.try_command(self.__run_ui_task_from_queue, self.root),
                    *(interval,)
                )
                break
            self.reportqueue.task_done()


def write_error_to_log():
    """Write the exception to the error log with a time stamp."""
    with open(os.path.join(sys.argv[1], ERROR_LOG), "a") as file:
        file.write(
            "".join(
                (
                    "\n\n\n",
                    " ".join(
                        (
                            APPLICATION_NAME,
                            "exception report at",
                            datetime.datetime.isoformat(
                                datetime.datetime.today()
                            ),
                        )
                    ),
                    "\n\n",
                    traceback.format_exc(),
                    "\n\n",
                )
            )
        )



# repertoiredbedit.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit toplevel to edit or insert repertoire record."""

from solentware_grid.gui.dataedit import DataEdit

from pgn_read.core.parser import PGN

from ..core.constants import TAG_OPENING
from .repertoiretoplevel import (
    RepertoireToplevel,
    RepertoireToplevelEdit,
)
from .toplevelpgn import EditPGNToplevel
from .constants import EMPTY_REPERTOIRE_GAME


class RepertoireDbEdit(EditPGNToplevel, DataEdit):
    """Edit PGN text for repertoire on database, or insert a new record.

    parent is used as the master argument in RepertoireToplevel calls.

    ui is used as the ui argument in RepertoireToplevel calls.

    newobject, parent, oldobject, and the one or two RepertoireToplevel
    instances created, are used as arguments in the super.__init__ call.

    showinitial determines whether a RepertoireToplevel is created for
    oldobject if there is one.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Attribute pgn_score_tags provides empty PGN tags to present when creating
    an insert Toplevel.  It is the empty PGN tags defined for repertoires in
    ChessTab..

    Attribute pgn_score_source provides the error key value to index a PGN
    game score with errors.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    pgn_score_name = "Repertoire"
    pgn_score_tags = EMPTY_REPERTOIRE_GAME
    pgn_score_source = "No opening name"

    def __init__(
        self,
        newobject=None,
        parent=None,
        oldobject=None,
        showinitial=True,
        ui=None,
    ):
        """Extend and create toplevel to edit or insert repertoire."""
        if not oldobject:
            showinitial = False
        super().__init__(
            newobject=newobject,
            parent=parent,
            oldobject=oldobject,
            newview=RepertoireToplevelEdit(master=parent, ui=ui),
            title="",
            oldview=RepertoireToplevel(master=parent, ui=ui)
            if showinitial
            else showinitial,
        )
        self._initialize()

    @property
    def ui_base_table(self):
        """Return the User Interface RepertoireGrid object."""
        return self.ui.base_repertoires

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.games_and_repertoires_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'editable' view."""
        return self.newview.ui

    def _set_item(self, view, object_):
        """Populate view with the repertoire extracted from object_."""
        self._set_default_source_for_object(object_)
        view.set_position_analysis_data_source()
        view.collected_game = next(
            PGN(game_class=view.gameclass).read_games(object_.get_srvalue())
        )
        view.set_and_tag_item_text()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a Repertoire object_.

        Default value of object_ is oldobject attribute from DataEdit class.

        """
        if object_ is None:
            object_ = self.oldobject
        if object_:
            try:
                return "  ".join(
                    (
                        self.pgn_score_name.join(("Edit ", ":")),
                        object_.value.collected_game.pgn_tags[TAG_OPENING],
                    )
                )
            except TypeError:
                return self.pgn_score_name.join(
                    ("Edit ", " - name unknown or invalid")
                )
            except KeyError:
                return self.pgn_score_name.join(
                    ("Edit ", " - name unknown or invalid")
                )
        else:
            return "".join(("Insert ", self.pgn_score_name))

    def _set_default_source_for_object(self, object_=None):
        """Set default source for Toplevel containing a Repertoire object_.

        Default value of object_ is oldobject attribute from DataEdit class.

        """
        if object_ is None:
            object_ = self.oldobject
        if object_ is not None:
            object_.value.set_game_source(self.pgn_score_source)



# chessndbmdu.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for dbm.ndbm."""

from solentware_base import ndbmdu_database

from ..shared.litedu import Litedu
from ..shared.alldu import chess_du, Alldu


class ChessndbmduError(Exception):
    """Exception class for chessndbmdu module."""


def chess_database_du(dbpath, *args, estimated_number_of_games=0, **kwargs):
    """Open database, import games and close database."""
    chess_du(ChessDatabase(dbpath, allowcreate=True), *args, **kwargs)

    # There are no recoverable file full conditions for dbm.ndbm (see DPT).
    return True


class ChessDatabase(Alldu, Litedu, ndbmdu_database.Database):
    """Provide custom deferred update for a database of games of chess."""

    def __init__(self, ndbmfile, **kargs):
        """Delegate with ChessndbmduError as exception class."""
        super().__init__(ndbmfile, ChessndbmduError, **kargs)



# setup.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)
"""chesstab setup file.

Retained so 'python setup.py ...' can be used by my build scripts until
conversion to 'python -m pip ...'.

All setup() arguments except dependency_links are set in setup.cfg file.
"""

from setuptools import setup
import configparser

if __name__ == "__main__":

    setup_args = {}

    # Derive dependency_links from install_requires for use by setup().
    # dependency_links is ignored by non-legacy pip commands; and these
    # packages have to be installed by separate commands quoting the URL
    # in the command line.  This code works because all version conditions
    # are '=='.
    parser = configparser.ConfigParser()
    parser.read("setup.cfg")
    for section in parser.sections():
        if section != "options":
            continue
        for key, value in parser.items(section):
            if key == "install_requires":
                install_requires = [v for v in value.splitlines() if v]
                setup_args["dependency_links"] = [
                    "-".join(required.split("==")).join(
                        ("http://solentware.co.uk/files/", ".tar.gz")
                    )
                    for required in install_requires
                ]
                break

    setup(**setup_args)



# topleveltext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which initialize and close record or filter rule displays.

The records are games, repertoires, and rule lists.  The rules are selection
rules, CQL statements, and run engine commands in Toplevels.

"""


class ToplevelText:
    """Provide default popup menu initialisation methods for record display."""

    binding_labels = ()

    # Without this, right-click, Shift F10, and Ctrl F10, in CQL* and Query*
    # toplevels cause AttributeError exceptions.  This just causes a null
    # popup menu to be posted: perhaps difficult to notice.  Maybe a reason to
    # show something is needed.
    # The Engine* classes do not have this problem, perhaps because of the
    # 'Run Engine' option or the absence of the 'main display' options.
    # The Game* and Repertoire* classes have reason for a popup here.
    def _create_widget_navigation_submenu_for_popup(self, popup):
        """Do nothing.

        Most ToplevelText subclasses require a handler for right-click,
        Shift F10, and Ctrl F10, events.

        A null popup menu is posted unless the subclass has a reason to
        implement a useful popup menu.
        """


class _ToplevelText:
    """Provide methods shared by this module's public _ToplevelText subclasses.

    The subclasses are DeleteText, EditText, and ShowText.
    """

    def _initialize(self):
        """Initialize the widgets and event bindings in Toplevel widget."""
        oldview = self.oldview
        self.ui_items_in_toplevels.add(oldview)
        self._set_item(oldview, self.object)
        self.parent.wm_title(self._get_title_for_object(self.object))
        self._initialize_item_bindings(oldview)

    def _initialize_item_bindings(self, item):
        """Initialize keypress and buttonpress bindings for item."""
        self.bind_buttons_to_widget(item.score)
        # item.set_score_pointer_to_score_bindings(False)

    def tidy_on_destroy(self):
        """Clear up after Toplevel destruction."""
        self.ui_items_in_toplevels.discard(self.oldview)
        self.ui_base_table.selection.clear()


class ShowText(_ToplevelText):
    """Show original text."""

    def dialog_ok(self):
        """Close the show record toplevel."""
        if self.ui.database is None:
            self.destroy_dialog_on_ok_and_blockchange()
            return False
        return super().dialog_ok()


class DeleteText(_ToplevelText):
    """Show original text for record deletion."""

    def dialog_ok(self):
        """Delete record and return delete action response (True for deleted).

        Check that database is open and is same one as deletion action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text="Cannot delete because not connected to a database"
            )
            self.destroy_dialog_on_ok_and_blockchange()
            return False
        # 'RuntimeError: dictionary changed size during iteration'
        # for deleting chess engine record first noticed 11 January 2023.
        # Does not affect deletion of other types of record.
        # Fixed by adjustments in .uci module to compensate for removal of
        # spare widget pool.  Root cause initially masked by RuntimeError
        # exception avoided by changes in solentware_grid.core.dataclient
        # refresh_widgets method.
        return super().dialog_ok()


class EditText(_ToplevelText):
    """Show original and editable text versions for record editing."""

    def _initialize(self):
        """Create widgets for the original and editable versions of text."""
        if self.oldview:
            self.ui_items_in_toplevels.add(self.oldview)
            self._set_item(self.oldview, self.oldobject)
            self._initialize_item_bindings(self.oldview)
        newview = self.newview
        self.ui_items_in_toplevels.add(newview)
        self._set_item(newview, self.newobject)
        self.parent.wm_title(self._get_title_for_object(self.newobject))
        self._initialize_item_bindings(newview)

    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text="Cannot update because not connected to a database"
            )
            self.destroy_dialog_on_ok_and_blockchange()
            return False
        return super().dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after Toplevel destruction."""
        self.ui_items_in_toplevels.discard(self.oldview)
        self.ui_items_in_toplevels.discard(self.newview)
        self.ui_base_table.selection.clear()



# enginetext.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to display a chess engine definition."""

import tkinter
import tkinter.messagebox
from urllib.parse import urlunsplit

from ..core.engine import Engine
from .eventspec import EventSpec
from .blanktext import NonTagBind, BlankText
from .sharedtext import SharedTextEngineText


class EngineText(SharedTextEngineText, BlankText):
    """Chess engine definition widget.

    panel is used as the panel argument for the super().__init__ call.

    ui is the user interface manager for an instance of EngineText, usually an
    instance of ChessUI.

    items_manager is used as the items_manager argument for the
    super().__init__ call.

    itemgrid is the ui reference to the DataGrid from which the record was
    selected.

    Subclasses are responsible for providing a geometry manager.

    Attribute _most_recent_bindings is set to indicate the initial set of
    event bindings.  Instances will override this as required.


    """

    def __init__(
        self, panel, ui=None, items_manager=None, itemgrid=None, **ka
    ):
        """Create widgets to display chess engine definition."""
        del itemgrid
        super().__init__(panel, items_manager=items_manager, **ka)
        self.ui = ui

        # Selection rule parser instance to process text.
        self.definition = Engine()

    # Engine description records are always shown in a Toplevel.
    # Dismiss item and database update actions by keypress and buttunpress
    # are assumed to be exposed by the associated solentware_misc class.
    def get_primary_activity_events(self):
        """Return tuple of game navigation keypresses and callbacks."""
        return ((EventSpec.databaseenginedisplay_run, self.run_engine),)

    def _set_primary_activity_bindings(self, switch=True):
        """Switch bindings for editing chess engine definition on or off."""
        self.set_event_bindings_score(
            self.get_primary_activity_events(), switch=switch
        )
        self.set_event_bindings_score(
            self.get_f10_popup_events(
                self._post_active_menu_at_top_left, self._post_active_menu
            ),
            switch=switch,
        )
        self.set_event_bindings_score(
            self._get_button_events(buttonpress3=self._post_active_menu),
            switch=switch,
        )

    def set_engine_definition(self, reset_undo=False):
        """Display the chess engine definition as text.

        reset_undo causes the undo redo stack to be cleared if True.  Set True
        on first display of an engine definition for editing so that repeated
        Ctrl-Z in text editing mode recovers the original engine definition.

        """
        if not self._is_text_editable:
            self.score.configure(state=tkinter.NORMAL)
        self.score.delete("1.0", tkinter.END)
        self._map_engine_definition()
        if self._most_recent_bindings != NonTagBind.NO_EDITABLE_TAGS:
            self._bind_for_primary_activity()
        if not self._is_text_editable:
            self.score.configure(state=tkinter.DISABLED)
        if reset_undo:
            self.score.edit_reset()

    def set_statusbar_text(self):
        """Set status bar to display chess engine definition name."""
        # self.ui.statusbar.set_status_text(self.definition.get_name_text())

    def get_name_engine_definition_dict(self):
        """Extract chess engine definition from Text widget."""
        engine = Engine()
        if engine.extract_engine_definition(
            self.score.get("1.0", tkinter.END).strip()
        ):
            return engine.__dict__
        return {}

    def _map_engine_definition(self):
        """Insert chess engine definition in Text widget.

        Method name arises from development history: the source class tags
        inserted text extensively and this method name survived the cull.

        """
        # No mapping of tokens to text in widget (yet).
        self.score.insert(
            tkinter.INSERT, self.definition.get_name_engine_command_text()
        )

    def run_engine(self, event=None):
        """Run chess engine."""
        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # if self.ui.uci.uci_drivers_reply is None:
        #    tkinter.messagebox.showinfo(
        #        parent=self.panel,
        #        title='Chesstab Restriction',
        #        message=' '.join(
        #            ('Starting an UCI chess engine is not allowed because',
        #             'an interface cannot be created:',
        #             'this is expected if running under Wine.')))
        #    return

        del event
        url = self.definition.engine_url_or_error_message()
        if isinstance(url, str):
            tkinter.messagebox.showerror(
                parent=self.panel, title="Run Engine", message=url
            )
            return
        if url.query:
            self.ui.run_engine(urlunsplit(url))
        elif url.path:
            command = url.path.split(" ", 1)
            if len(command) == 1:
                self.ui.run_engine(command[0])
            else:
                self.ui.run_engine(command[0], args=command[1].strip())
        else:
            tkinter.messagebox.showerror(
                parent=self.panel,
                title="Run Engine",
                message="".join(
                    (
                        "Unable to run engine for\n\n",
                        self.definition.get_engine_command_text(),
                    )
                ),
            )



# __init__.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a Berkeley DB database of chess games using tkinter Tcl interface.

This interface can be used if installed Berkeley DB is built with Tcl support.
"""



# repertoiretoplevel.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Toplevel widgets to display and edit repertoires.

These two classes display repertoires in their own Toplevel widget: they are
used in the repertoiredbdelete, repertoiredbedit, and repertoiredbshow,
modules.

"""
from .repertoire import Repertoire
from .repertoireedit import RepertoireEdit
from .toplevelpgn import ToplevelPGN


class RepertoireToplevel(ToplevelPGN, Repertoire):
    """Customize Repertoire seen in a ToplevelPGN widget."""


class RepertoireToplevelEdit(ToplevelPGN, RepertoireEdit):
    """Customize RepertoireEdit seen in a ToplevelPGN widget."""

    def _create_primary_activity_popup(self):
        """Create popup menu for a repertoire widget."""
        popup = super()._create_primary_activity_popup()
        self._add_pgn_navigation_to_submenu_of_popup(
            popup, index=self.analyse_popup_label
        )
        self._add_pgn_insert_to_submenu_of_popup(
            popup,
            include_ooo=True,
            include_move_rav=True,
            index=self.analyse_popup_label,
        )
        return popup



# export_chessql.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess chessql (partial position) exporters."""

from . import chessrecord, filespec
from .cqlstatement import CQLStatement


def export_all_positions(database, filename):
    """Export CQL statements in database to text file in internal format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordPartial()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.PARTIAL_FILE_DEF, filespec.PARTIAL_FILE_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                gamesout.write(instance.get_srvalue())
                gamesout.write("\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return True


def export_selected_positions(grid, filename):
    """Export CQL statements in grid to textfile."""
    if filename is None:
        return
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        instance = chessrecord.ChessDBrecordPartial()
        instance.set_database(database)
        with open(filename, "w") as gamesout:
            for bookmark in sorted(grid.bookmarks):
                instance.load_record(
                    database.get_primary_record(
                        filespec.PARTIAL_FILE_DEF, bookmark[0]
                    )
                )
                gamesout.write(instance.get_srvalue())
                gamesout.write("\n")
        return
    database = grid.get_data_source().dbhome
    instance = chessrecord.ChessDBrecordPartial()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.PARTIAL_FILE_DEF, filespec.PARTIAL_FILE_DEF
    )
    try:
        with open(filename, "w") as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                gamesout.write(instance.get_srvalue())
                gamesout.write("\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return


def export_single_position(partialposition, filename):
    """Export CQL statement to textfile."""
    if filename is None:
        return
    cql_statement = CQLStatement()
    cql_statement.process_statement(partialposition)
    if not cql_statement.is_statement():
        return
    with open(filename, "w") as gamesout:
        gamesout.write(cql_statement.get_statement_text())



# chessdptdu.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using DPT single-step deferred update.

This module on Windows only.  Use multi-step module on Wine because Wine
support for a critical function used by single-step is not reliable. There
is no sure way to spot that module is running on Wine.

See www.dptoolkit.com for details of DPT

"""

import os
from io import StringIO
import multiprocessing  # Removed later by 'del multiprocessing'.

import tkinter
import tkinter.messagebox

# pylint will always give import-error message on non-Microsoft Windows
# systems.
# Wine counts as a Microsft Windows system.
# It is reasonable to not install 'dptdb.dptapi'.
# The importlib module is used to import chessdptdu if needed.
from dptdb.dptapi import FISTAT_DEFERRED_UPDATES

from solentware_base import dptdu_database
from solentware_base.core.constants import (
    FILEDESC,
    BRECPPG,
    TABLE_B_SIZE,
)

from ..core.filespec import (
    FileSpec,
    GAMES_FILE_DEF,
    PIECES_PER_POSITION,
    POSITIONS_PER_GAME,
)
from ..core.chessrecord import ChessDBrecordGameImport
from ..shared.archivedu import Archivedu

# Current practical way to determine if running on Wine, taking advantage of
# a problem found in ..core.uci which prevents UCI chess engines being used to
# do position analysis in this environment.  Here the consequence is Wine's
# useless answers to queries about memory usage made within DPT and the impact
# on the useful frequency of progress reports.
# CHUNKGAMES deals this this under Wine.  A safe value for a PC with 2Gb memory
# is 6144.  Almost get away with 8192 but it seems system housekeeping tasks,
# if nothing else, will cause this to fail at some point during a large import.
# Must be safe to add 8192 for every extra 2Gb, and possibly every extra 1Gb
# assuming system tasks are not greedy with the extra memory.
# The DPT segment size is 65280 because 32 bytes are reserved and 8160 bytes of
# the 8192 byte page are used for the bitmap.  Use 65536 as segment size and
# ignore the DPT segment size so report points are similar to other database
# engines.
# DB_SEGMENT_SIZE has no effect on processing apart from report points.
DB_SEGMENT_SIZE = 65536
try:
    multiprocessing.Queue()
    _DEFERRED_UPDATE_POINTS = (DB_SEGMENT_SIZE - 1,)
except OSError:
    _DEFERRED_UPDATE_POINTS = tuple(
        i
        for i in range(
            DB_SEGMENT_SIZE // 8 - 1, DB_SEGMENT_SIZE, DB_SEGMENT_SIZE // 8
        )
    )
del multiprocessing
del DB_SEGMENT_SIZE

CHUNKGAMES = 6144


class ChessdptduError(Exception):
    """Exception class for chessdptdu module."""


def chess_dptdu(
    dbpath,
    pgnpaths,
    file_records=None,
    reporter=lambda text, timestamp=True: None,
    estimated_number_of_games=0,
):
    """Open database, import games and close database."""
    cdb = ChessDatabase(dbpath, allowcreate=True)
    importer = ChessDBrecordGameImport()
    cdb.open_database(files=file_records)
    for pgnfile in pgnpaths:
        with open(pgnfile, "r", encoding="iso-8859-1") as source:
            importer.import_pgn(cdb, source, pgnfile, reporter=reporter)
    reporter("Finishing import: please wait.")
    reporter("", timestamp=False)
    cdb.close_database_contexts(files=file_records)
    cdb.open_database_contexts(files=file_records)
    status = True
    for file in (
        cdb.specification.keys() if file_records is None else file_records
    ):
        if (
            FISTAT_DEFERRED_UPDATES
            != cdb.table[file].get_file_parameters(cdb.dbenv)["FISTAT"][0]
        ):
            status = False
    cdb.close_database_contexts()
    return status


def chess_dptdu_chunks(
    dbpath,
    pgnpaths,
    file_records,
    reporter=lambda text, timestamp=True: None,
    estimated_number_of_games=0,
):
    """Open database, import games in fixed chunks and close database."""

    def write_chunk(pgnpath, sample=None):
        cdb.open_database(files=sample)

        # chess_dptdu_chunks does the reporting so do not pass reporter to
        # import_pgn, but the name of the source file is mandantory.
        importer.import_pgn(cdb, StringIO("".join(games)), pgnpath)

        cdb.close_database_contexts()
        games[:] = []

    cdb = ChessDatabase(dbpath, allowcreate=True)
    importer = ChessDBrecordGameImport()
    games = []
    gamelines = []
    newgamestring = '[Event "'
    records = file_records
    for pgnfile in pgnpaths:
        with open(pgnfile, "r", encoding="iso-8859-1") as inp:
            line = inp.readline()
            reporter("Extracting games from " + pgnfile)
            count = 0
            while line:
                if line.startswith(newgamestring):
                    games.append("".join(gamelines))
                    gamelines = []
                    if len(games) >= CHUNKGAMES:
                        write_chunk(pgnfile, sample=records)
                        count += 1
                        reporter(
                            " ".join(
                                (
                                    "Chunk",
                                    str(count),
                                    "written, total games added:",
                                    str(CHUNKGAMES * count),
                                )
                            )
                        )
                gamelines.append(line)
                line = inp.readline()
        if gamelines:
            games.append("".join(gamelines))
            write_chunk(pgnfile, sample=records)
        reporter("Extraction from " + pgnfile + " done")
        reporter("", timestamp=False)
    reporter("Finishing import: please wait.")
    reporter("", timestamp=False)
    cdb.open_database_contexts(files=file_records)
    status = True
    for file in file_records:
        if (
            FISTAT_DEFERRED_UPDATES
            != cdb.table[file].get_file_parameters(cdb.dbenv)["FISTAT"][0]
        ):
            status = False
    cdb.close_database_contexts()
    return status


class ChessDatabaseDeferred(dptdu_database.Database):
    """Provide deferred update methods for a database of games of chess.

    Subclasses must include a subclass of dptbase.Database as a superclass.

    """

    # ChessDatabaseDeferred.deferred_update_points is not needed in DPT, like
    # the similar attribute in chessdbbitdu.ChessDatabase for example, because
    # DPT does it's own memory management for deferred updates.
    # The same attribute is provided to allow the import_pgn method called in
    # this module's chess_dptdu and chess_dptdu_chunks functions to report
    # progress at regular intervals.
    # The values are set differently because Wine does not give a useful answer
    # to DPT's memory usage questions.
    deferred_update_points = frozenset(_DEFERRED_UPDATE_POINTS)

    def __init__(
        self,
        databasefolder,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        ddnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )
        # Deferred update for games file only
        for ddname in list(ddnames.keys()):
            if ddname != GAMES_FILE_DEF:
                del ddnames[ddname]

        if not kargs.get("allowcreate", False):
            try:
                for ddname in ddnames:
                    if FILEDESC in ddnames[ddname]:
                        del ddnames[ddname][FILEDESC]
            except Exception as error:
                if __name__ == "__main__":
                    raise
                raise ChessdptduError("DPT description invalid") from error

        try:
            super().__init__(ddnames, databasefolder, **kargs)
        except ChessdptduError as error:
            if __name__ == "__main__":
                raise
            raise ChessdptduError("DPT description invalid") from error

        # Retain import estimates for increase size by button actions
        self._import_estimates = None
        self._notional_record_counts = None
        # Methods passed by UI to populate report widgets
        self._reporter = None

    def open_database(self, files=None):
        """Delegate then raise ChessdptduError if database is inconsistent.

        Normally return None with the database open.

        Close the database and raise ChessdptduError exception if the
        database FISTAT parameter is not equal FISTAT_DEFERRED_UPDATES.

        """
        super().open_database(files=files)
        viewer = self.dbenv.Core().GetViewerResetter()
        for dbo in self.table.values():
            if (
                viewer.ViewAsInt("FISTAT", dbo.opencontext)
                != FISTAT_DEFERRED_UPDATES
            ):
                break
        else:
            if files is None:
                files = dict()
            self.increase_database_size(files=files)
            return
        self.close_database()
        raise ChessdptduError("A file is not in deferred update mode")

    def open_context_prepare_import(self):
        """Open all files normally."""
        super().open_database()

    def get_archive_names(self, files=()):
        """Return specified files and existing operating system files."""
        specs = {f for f in files if f in self.table}
        names = [v.file for k, v in self.table.items() if k in specs]
        exists = [
            os.path.basename(n)
            for n in names
            if os.path.exists(".".join((n, "bz2")))
        ]
        return (names, exists)

    def _get_pages_for_record_counts(self, counts=(0, 0)):
        """Return Table B and Table D pages needed for record counts."""
        brecppg = self.table[GAMES_FILE_DEF].filedesc[BRECPPG]
        return (
            counts[0] // brecppg,
            (counts[1] * self.table[GAMES_FILE_DEF].btod_factor) // brecppg,
        )

    def _get_database_table_sizes(self, files=None):
        """Return Table B and D size and usage in pages for files."""
        if files is None:
            files = dict()
        filesize = dict()
        for key, value in self.get_database_parameters(
            files=list(files.keys())
        ).items():
            filesize[key] = (
                value["BSIZE"],
                value["BHIGHPG"],
                value["DSIZE"],
                value["DPGSUSED"],
            )
        increase = self.get_database_increase(files=files)
        self.close_database_contexts()
        return filesize, increase

    def add_import_buttons(
        self, master, try_command_wrapper, try_event_wrapper, bind, widget
    ):
        """Add button actions for DPT to Import dialogue.

        Increase data and index space available.

        """
        index = tkinter.Button(
            master=master,
            text="Increase Index",
            underline=13,
            command=try_command_wrapper(self._increase_index, master),
        )
        index.pack(side=tkinter.RIGHT, padx=12)
        bind(widget, "<Alt-x>", try_event_wrapper(self._increase_index))
        data = tkinter.Button(
            master=master,
            text="Increase Data",
            underline=9,
            command=try_command_wrapper(self._increase_data, master),
        )
        data.pack(side=tkinter.RIGHT, padx=12)
        bind(widget, "<Alt-d>", try_event_wrapper(self._increase_data))

    def _increase_data(self, event=None):
        """Add maximum of current free space and default size to Table B.

        event is ignored and is present for compatibility between button click
        and keypress,

        """
        del event
        self.open_database_contexts(files=(GAMES_FILE_DEF,))
        increase_done = False
        for key, value in self.get_database_parameters(
            files=(GAMES_FILE_DEF,)
        ).items():
            bsize = value["BSIZE"]
            bused = max(0, value["BHIGHPG"])
            bneeded = self._get_pages_for_record_counts(
                self._notional_record_counts[key]
            )[0]
            bincrease = min(bneeded * 2, bsize - bused)
            message = "".join(
                (
                    "The free data size of the ",
                    key,
                    " file will be increased from ",
                    str(bsize - bused),
                    " pages to ",
                    str(bincrease + bsize - bused),
                    " pages.",
                )
            )
            if len(self.table[key].get_extents()) % 2 == 0:
                message = "".join(
                    (
                        message,
                        "\n\nAt present it is better to do index increases ",
                        "first for this file, if you need to do any, because ",
                        "a new extent (fragment) would not be needed.",
                    )
                )
            if tkinter.messagebox.askyesno(
                title="Increase Data Size",
                message="".join(
                    (
                        message,
                        "\n\nDo you want to increase the data size?",
                    )
                ),
            ):
                increase_done = True
                self.table[key].opencontext.Increase(bincrease, False)
        if increase_done:
            self._reporter.append_text(
                " ".join(
                    (
                        "Recalculation of planned database size increases",
                        "after data size increase by user action.",
                    )
                )
            )
            self._reporter.append_text_only("")
            self._report_plans_for_estimate()
        self.close_database_contexts()

    def _increase_index(self, event=None):
        """Add maximum of current free space and default size to Table D.

        event is ignored and is present for compatibility between button click
        and keypress,

        """
        del event
        self.open_database_contexts(files=(GAMES_FILE_DEF,))
        increase_done = False
        for key, value in self.get_database_parameters(
            files=(GAMES_FILE_DEF,)
        ).items():
            dsize = value["DSIZE"]
            dused = value["DPGSUSED"]
            dneeded = self._get_pages_for_record_counts(
                self._notional_record_counts[key]
            )[1]
            dincrease = min(dneeded * 2, dsize - dused)
            message = "".join(
                (
                    "The free index size of the ",
                    key,
                    " file will be increased from ",
                    str(dsize - dused),
                    " pages to ",
                    str(dincrease + dsize - dused),
                    " pages.",
                )
            )
            if len(self.table[key].get_extents()) % 2 != 0:
                message = "".join(
                    (
                        message,
                        "\n\nAt present it is better to do data increases ",
                        "first for this file, if you need to do any, because ",
                        "a new extent (fragment) would not be needed.",
                    )
                )
            if tkinter.messagebox.askyesno(
                title="Increase Index Size",
                message="".join(
                    (
                        message,
                        "\n\nDo you want to increase the index size?",
                    )
                ),
            ):
                increase_done = True
                self.table[key].opencontext.Increase(dincrease, True)
        if increase_done:
            self._reporter.append_text(
                " ".join(
                    (
                        "Recalculation of planned database size increases",
                        "after index size increase by user action.",
                    )
                )
            )
            self._reporter.append_text_only("")
            self._report_plans_for_estimate()
        self.close_database_contexts()

    def get_file_sizes(self):
        """Return dictionary of notional record counts for data and index."""
        return self._notional_record_counts

    def report_plans_for_estimate(self, estimates, reporter):
        """Calculate and report file size adjustments to do import.

        Note the reporter and headline methods for initial report and possible
        later recalculations.

        Pass estimates through to self._report_plans_for_estimate

        """
        # See comment near end of class definition Chess in relative module
        # ..gui.chess for explanation of this change.
        self._reporter = reporter
        self._report_plans_for_estimate(estimates=estimates)

    def _report_plans_for_estimate(self, estimates=None):
        """Recalculate and report file size adjustments to do import.

        Create dictionary of effective game counts for sizing Games file.
        This will be passed to the import job which will increase Table B and
        Table D according to file specification.

        The counts for Table B and Table D can be different.  If the average
        data bytes per game is greater than Page size / Records per page the
        count must be increased to allow for the unused record numbers.  If
        the average positions per game or pieces per position are not the
        values used to calculate the steady-state ratio of Table B to Table D
        the count must be adjusted to compensate.

        """
        append_text = self._reporter.append_text
        append_text_only = self._reporter.append_text_only
        if estimates is not None:
            self._import_estimates = estimates
        (
            gamecount,
            bytes_per_game,
            positions_per_game,
            pieces_per_game,
        ) = self._import_estimates[:4]
        brecppg = self.table[GAMES_FILE_DEF].filedesc[BRECPPG]
        d_count = (gamecount * (positions_per_game + pieces_per_game)) // (
            POSITIONS_PER_GAME * (1 + PIECES_PER_POSITION)
        )
        if bytes_per_game > (TABLE_B_SIZE // brecppg):
            b_count = int(
                (gamecount * bytes_per_game) / (TABLE_B_SIZE / brecppg)
            )
        else:
            b_count = gamecount
        self._notional_record_counts = {
            GAMES_FILE_DEF: (b_count, d_count),
        }
        append_text("Current file size and free space:")
        free = dict()
        sizes, increases = self._get_database_table_sizes(
            files=self._notional_record_counts
        )
        for filename, bdsize in sizes.items():
            bsize, bused, dsize, dused = bdsize
            bused = max(0, bused)
            free[filename] = (bsize - bused, dsize - dused)
            append_text_only(filename)
            append_text_only(
                " ".join(("Current data area size", str(bsize), "pages"))
            )
            append_text_only(
                " ".join(("Current index area size", str(dsize), "pages"))
            )
            append_text_only(
                " ".join(
                    ("Current data area free", str(bsize - bused), "pages")
                )
            )
            append_text_only(
                " ".join(
                    ("Current index area free", str(dsize - dused), "pages")
                )
            )
        append_text_only("")
        append_text("File space needed for import:")
        for filename, nr_count in self._notional_record_counts.items():
            append_text_only(filename)
            b_pages, d_pages = self._get_pages_for_record_counts(nr_count)
            append_text_only(
                " ".join(("Estimated", str(b_pages), "pages needed for data"))
            )
            append_text_only(
                " ".join(
                    ("Estimated", str(d_pages), "pages needed for indexes")
                )
            )
        append_text_only("")
        append_text("File size increases planned and free space when done:")
        for filename, increments in increases.items():
            b_incr, d_incr = increments
            b_free, d_free = free[filename]
            append_text_only(filename)
            append_text_only(
                " ".join(("Data area increase", str(b_incr), "pages"))
            )
            append_text_only(
                " ".join(("Index area increase", str(d_incr), "pages"))
            )
            append_text_only(
                " ".join(("Data area free", str(b_incr + b_free), "pages"))
            )
            append_text_only(
                " ".join(("Index area free", str(d_incr + d_free), "pages"))
            )
        append_text_only("")
        append_text_only(
            "".join(
                (
                    "Comparison of the required and free data or index ",
                    "space may justify using the Increase Data and, or, ",
                    "Increase Index actions to get more space immediately ",
                    "given your knowledge of the PGN file being imported.",
                )
            )
        )
        append_text_only("")
        append_text_only("")


class ChessDatabase(Archivedu, ChessDatabaseDeferred):
    """Provide single-step deferred update for a database of games of chess."""


if __name__ == "__main__":

    import tkinter.ttk
    import tkinter.filedialog

    from solentware_bind.gui.bindings import Bindings

    binder = Bindings()
    root = tkinter.Tk()
    root.wm_title("Sample do_deferred_updates")
    frame = tkinter.ttk.Frame(master=root)
    text = tkinter.Text(master=frame, wrap=tkinter.WORD)
    frame.pack(fill=tkinter.Y, expand=tkinter.TRUE)
    text.pack(fill=tkinter.BOTH, expand=tkinter.TRUE)
    databasepath = tkinter.filedialog.askdirectory(
        parent=root, title="Select database directory", initialdir="~"
    )
    if databasepath:
        text.insert(tkinter.END, "Database directory: " + databasepath + "\n")
    else:
        text.insert(tkinter.END, "No database selected\n")

    def quit_(*a):
        """Destroy application."""
        del a
        root.destroy()

    def show_menu(event=None):
        """Post popup menu at pointer location."""
        menu.tk_popup(*event.widget.winfo_pointerxy())

    menu = tkinter.Menu(master=root, tearoff=False)
    if databasepath:

        class _FilePath:
            """Wrap filepath as a class attribute.

            pylint by default reports problems with filepath as a module
            constant and as a global in the proceed() method.

            Class attributes are allowed any format by pylint by default.
            """

            filepath = None

        def do_function(dbp, pgnpath):
            """Run chess_dptdu with database at dbp and PGN file at fp."""
            chess_dptdu(dbp, pgnpath)
            text.insert(tkinter.END, "Process finished\nProceed or Quit?\n\n")

        def proceed(*a):
            """Run a process by the do_function function.

            *a soaks up any arguments: in particular event if proceed is
            invoked by KeyPress rather than a menu command.

            """
            del a
            filepath = _FilePath.filepath  # Used to be 'global filepath'.
            if filepath:
                text.insert(tkinter.END, "\nProcess started, please wait.\n")
                text.after(1, do_function, *(databasepath, [filepath]))
                _FilePath.filepath = None
            elif databasepath:
                filepath = tkinter.filedialog.askopenfilename(
                    parent=root, title="Select data file", initialdir="~"
                )
                _FilePath.filepath = filepath
                if filepath:
                    text.insert(tkinter.END, "Data file: " + filepath + "\n")
                else:
                    text.insert(tkinter.END, "No data file selected\n")

        menu.add_separator()
        menu.add_command(
            label="Proceed", command=proceed, accelerator="Alt F2"
        )
        menu.add_command(label="Quit", command=quit_, accelerator="Alt F11")
        menu.add_separator()
        text.insert(
            tkinter.END, "\nRight-click to proceed after each step\n\n"
        )
        binder.bind(text, "<Alt-KeyPress-F2>", function=proceed)
        binder.bind(text, "<Alt-KeyPress-F11>", function=quit_)
        binder.bind(text, "<ButtonPress-3>", function=show_menu)
    else:
        menu.add_separator()
        menu.add_command(label="Quit", command=quit_, accelerator="Alt F11")
        menu.add_separator()
        text.insert(tkinter.END, "\nRight-click to quit\n")
        binder.bind(text, "<Alt-KeyPress-F11>", function=quit_)
        binder.bind(text, "<ButtonPress-3>", function=show_menu)
    root.mainloop()



# chesslmdbdu.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for Symas LMMD."""

from solentware_base import lmdbdu_database
from solentware_base.core.constants import DEFAULT_MAP_PAGES

from ..shared.lmdbdudb import LmdbduDb
from ..shared.lmdbdu import Lmdbdu
from ..shared.alldu import chess_du, Alldu
from ..core import filespec


class ChesslmdbduError(Exception):
    """Exception class for chesslmdbdu module."""


def chess_database_du(dbpath, *args, estimated_number_of_games=0, **kwargs):
    """Open database, import games and close database."""
    chess_du(
        ChessDatabase(
            dbpath,
            allowcreate=True,
            estimated_number_of_games=estimated_number_of_games,
        ),
        *args,
        **kwargs,
    )

    # There are no recoverable file full conditions for Symas LMMD (see DPT).
    return True


class ChessDatabase(Alldu, LmdbduDb, Lmdbdu, lmdbdu_database.Database):
    """Provide custom deferred update for a database of games of chess."""

    def __init__(self, DBfile, estimated_number_of_games=0, **kargs):
        """Delegate with ChesslmdbduError as exception class."""
        super().__init__(
            DBfile,
            ChesslmdbduError,
            **kargs,
        )
        self.open_database()
        (
            map_bytes, map_pages, used_bytes, used_pages, stats
        ) = self.database_stats_summary()
        self.close_database()
        mfpas = filespec.LMMD_MINIMUM_FREE_PAGES_AT_START
        if map_pages - used_pages < mfpas:
            short = mfpas + map_pages - used_pages
        else:
            short = 0
        self.map_blocks = (
            1
            + (map_pages + short) // DEFAULT_MAP_PAGES
            + estimated_number_of_games // 1000
        )



# lmdbdudb.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The LmdbduDb class for lmdb interface to Symas LMMD.

This module uses the solentware_base.core._lmdb Database class.

"""

from solentware_base.core._lmdb import Database

from .dptcompatdu import DptCompatdu


class LmdbduDb(DptCompatdu, Database):
    """Provide deferred update methods shared by Symas LMMD interfaces."""



# dptduchunk.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file in chunks using DPT single-step deferred update.

The chunks are are chosen so the job does not run out of memory when running
under wine.  The test for low memory usually does not work (if I recall
correctly there was one version of wine where a test, not the one used before
or since, did spot the condition allowing DPT to flush it's buffers in time).
"""


if __name__ == "__main__":

    import tkinter

    import tkinter.messagebox
    import tkinter.filedialog

    from chesstab.dpt.chessdptdu import chess_dptdu_chunks

    root = tkinter.Tk()
    root.wm_title(string="Test Import Chess Games")
    root.wm_iconify()
    dbdir = tkinter.filedialog.askdirectory(title="Open Chess database folder")
    if dbdir:
        filename = tkinter.filedialog.askopenfilename(
            title="PGN file of Games",
            defaultextension=".pgn",
            filetypes=(("PGN Chess Games", "*.pgn"),),
        )
        if filename:
            if tkinter.messagebox.askyesno(
                title="Import Games", message="Proceed with import"
            ):
                chess_dptdu_chunks(dbdir, (filename,), {})
    root.destroy()



# colourscheme.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""A chess board colour scheme chooser."""

import tkinter
import tkinter.font

from solentware_misc.gui.colourslider import ColourSlider

from solentware_bind.gui.bindings import Bindings

from pgn_read.core.parser import PGN

from . import constants, fonts, options
from . import game, board


class _ColourScheme(Bindings):
    """Widgets to select fonts and colours for chess board and game score.

    Use the subclasses defined in this module depending on what fits on
    the screen.
    """

    restore_focus = None
    game = None
    b_families = None
    b_family = None
    b_weight = None
    s_families = None
    s_family = None
    s_weight = None
    s_slant = None
    s_size = None

    def __init__(self, ui=None, height=500, title="Style", **kargs):
        """Create Toplevel and fonts to be modified.

        ui is the ChessUI instance
        height - passed to Tkinter.Toplevel.
        title - use is w.wm_title(title).

        """
        del kargs
        super().__init__()
        self.ui = ui
        tags = ui.tags_variations_comments_font
        moves = ui.moves_played_in_game_font
        pieces = ui.boardfont
        self._ok = False
        if tags:
            self.tags_variations_comments_font = tags.copy()
        else:
            self.tags_variations_comments_font = tkinter.font.nametofont(
                constants.TAGS_VARIATIONS_COMMENTS_FONT
            ).copy()
        if moves:
            self.moves_played_in_game_font = moves.copy()
        else:
            self.moves_played_in_game_font = tkinter.font.nametofont(
                constants.MOVES_PLAYED_IN_GAME_FONT
            ).copy()
        if pieces:
            self.boardfont = pieces.copy()
        else:
            self.boardfont = tkinter.font.nametofont(
                constants.PIECES_ON_BOARD_FONT
            ).copy()
        # Wildpieces font is based on moves font but managed like boardfont
        if moves:
            self.wildpiecesfont = moves.copy()
        else:
            self.wildpiecesfont = tkinter.font.nametofont(
                constants.WILDPIECES_ON_BOARD_FONT
            ).copy()
        self.chooser = tkinter.Toplevel(cnf=dict(height=height, width=700))
        self.chooser.wm_resizable(False, False)
        self.chooser.wm_title(title)
        self.chooser.pack_propagate(False)

    def _apply_modified_font(self, modifiedfont, font):
        """Copy modifiedfont attributes to font."""
        if self.is_ok():
            fonts.copy_font_attributes(modifiedfont.actual(), font)

    def _apply_modified_board_font(self, modifiedfont, font):
        """Copy modifiedfont attributes to board font."""
        if self.is_ok():
            fonts.copy_board_font_attributes(modifiedfont.actual(), font)

    def apply_pieces_on_board_font_to_font(self, font):
        """Copy self.boardfont attributes to board font."""
        self._apply_modified_board_font(self.boardfont, font)

    def apply_to_named_fonts(self):
        """Apply modified fonts to all displayed games."""
        self._apply_to_named_font_moves_played_in_game()
        self._apply_to_named_font_pieces_on_board()
        self._apply_to_named_font_wildpieces_on_board()
        self._apply_to_named_font_tags_variations_comments()

    def _apply_to_named_font_moves_played_in_game(self):
        """Apply moves played font to all displayed games."""
        self._apply_modified_font(
            self.moves_played_in_game_font,
            tkinter.font.nametofont(constants.MOVES_PLAYED_IN_GAME_FONT),
        )

    def _apply_to_named_font_pieces_on_board(self):
        """Apply pieces font to all displayed games."""
        self._apply_modified_board_font(
            self.boardfont,
            tkinter.font.nametofont(constants.PIECES_ON_BOARD_FONT),
        )

    def _apply_to_named_font_tags_variations_comments(self):
        """Apply variations and comments font to all displayed games."""
        self._apply_modified_font(
            self.tags_variations_comments_font,
            tkinter.font.nametofont(constants.TAGS_VARIATIONS_COMMENTS_FONT),
        )

    def _apply_to_named_font_wildpieces_on_board(self):
        """Apply wildpieces font to all displayed games."""
        self._apply_modified_board_font(
            self.wildpiecesfont,
            tkinter.font.nametofont(constants.WILDPIECES_ON_BOARD_FONT),
        )

    def _get_button_definitions(self):
        """Return button definitions."""
        return (
            ("OK", "OK button Tooltip.", True, -1, self._on_ok),
            ("Cancel", "Cancel button Tooltip.", True, 2, self._on_cancel),
        )

    def _create_buttons(self, buttons):
        """Create buttons."""
        buttons_frame = tkinter.Frame(master=self.chooser)
        buttons_frame.pack(side=tkinter.BOTTOM, fill=tkinter.X)
        buttonrow = buttons_frame.pack_info()["side"] in ("top", "bottom")
        for index, item in enumerate(buttons):
            button = tkinter.Button(
                master=buttons_frame,
                text=item[0],
                underline=item[3],
                command=self.try_command(item[4], buttons_frame),
            )
            if buttonrow:
                buttons_frame.grid_columnconfigure(index * 2, weight=1)
                button.grid_configure(column=index * 2 + 1, row=0)
            else:
                buttons_frame.grid_rowconfigure(index * 2, weight=1)
                button.grid_configure(row=index * 2 + 1, column=0)
        if buttonrow:
            buttons_frame.grid_columnconfigure(len(buttons * 2), weight=1)
        else:
            buttons_frame.grid_rowconfigure(len(buttons * 2), weight=1)

    def _create_colour_frame(self):
        """Create colour selector widgets."""
        colour_frame = tkinter.Frame(master=self.chooser)
        colour_frame.pack(fill=tkinter.X)
        gboard = self.game.board
        for row, label, color, setter in (
            (
                0,
                "Light squares",
                gboard.litecolor,
                self._set_lite_square_bgcolour,
            ),
            (
                1,
                "Dark squares",
                gboard.darkcolor,
                self._set_default_source_for_object,
            ),
            (
                2,
                "White pieces",
                gboard.whitecolor,
                self._set_white_piece_fgcolour,
            ),
            (
                3,
                "Black pieces",
                gboard.blackcolor,
                self._set_black_piece_fgcolour,
            ),
            (4, "Rest of line", self.game.l_color, self._set_tag_rest_of_line),
            (5, "Move", self.game.m_color, self._set_tag_move),
            (
                6,
                "Alternatives",
                self.game.am_color,
                self._set_tag_alternatives,
            ),
            (
                7,
                "Start of line",
                self.game.v_color,
                self._set_tag_start_of_line,
            ),
        ):
            ChessColourSlider(
                colour_frame,
                row=row,
                colour=color,
                label=label,
                tag_onfigure_bachground_colour=setter,
            )

    def _create_font_frame(self):
        """Create font selection widget."""

        def focus(widget):
            def fset(event=None):
                del event
                widget.focus_set()

            return fset

        font_frame = tkinter.Frame(master=self.chooser)
        font_frame.pack(fill=tkinter.X)
        # font chooser for board
        font_frame.columnconfigure(0, weight=1, uniform="fontpanels")
        tkinter.Label(master=font_frame, text="Pieces").grid_configure(
            row=0, column=0
        )
        fontfr = tkinter.Frame(master=font_frame, cnf=dict(padx=5))
        boardfr = tkinter.Frame(master=fontfr)
        self.b_families = tkinter.Listbox(boardfr)
        scrollfont = tkinter.Scrollbar(boardfr)
        scrollfont.configure(
            command=self.try_command(self.b_families.yview, scrollfont)
        )
        self.b_families.configure(yscrollcommand=scrollfont.set)
        for family in sorted(tkinter.font.families()):
            self.b_families.insert(tkinter.END, family)
        self.b_families.pack(
            side=tkinter.LEFT, expand=tkinter.TRUE, fill=tkinter.Y
        )
        scrollfont.pack(side=tkinter.LEFT, fill=tkinter.Y)
        boardfr.pack(fill=tkinter.Y, expand=tkinter.TRUE)
        wssf = tkinter.Frame(master=fontfr)
        self.b_family = tkinter.Label(
            master=wssf, text=self.boardfont["family"]
        )
        self.b_family.grid_configure(column=0, row=1, sticky=tkinter.EW)
        self.b_weight = tkinter.IntVar()
        checkbutton = tkinter.Checkbutton(
            master=wssf,
            text="Bold",
            variable=self.b_weight,
            command=self.try_command(self._set_board_font_weight, wssf),
            indicatoron=tkinter.FALSE,
        )
        checkbutton.grid_configure(column=0, row=3, sticky=tkinter.EW)
        self.bind(
            checkbutton,
            "<ButtonPress>",
            function=self.try_event(focus(checkbutton)),
        )
        self.b_weight.set(tkinter.TRUE)
        wssf.grid_columnconfigure(0, weight=1)
        wssf.grid_rowconfigure(0, minsize=5)
        wssf.grid_rowconfigure(2, minsize=5)
        wssf.grid_rowconfigure(4, minsize=5)
        wssf.pack(fill=tkinter.X)
        fontfr.grid_configure(row=1, column=0, sticky=tkinter.NS)
        self.bind(
            self.b_families,
            "<<ListboxSelect>>",
            function=self.try_event(self._set_board_font_family),
        )
        # font chooser for score
        font_frame.columnconfigure(1, weight=2, uniform="fontpanels")
        tkinter.Label(master=font_frame, text="Score").grid_configure(
            row=0, column=1
        )
        fontfr = tkinter.Frame(master=font_frame, cnf=dict(padx=5))
        scorefr = tkinter.Frame(master=fontfr)
        self.s_families = tkinter.Listbox(scorefr)
        scrollfont = tkinter.Scrollbar(scorefr)
        scrollfont.configure(
            command=self.try_command(self.s_families.yview, scrollfont)
        )
        self.s_families.configure(yscrollcommand=scrollfont.set)
        for family in sorted(tkinter.font.families()):
            self.s_families.insert(tkinter.END, family)
        self.s_families.pack(
            side=tkinter.LEFT, expand=tkinter.TRUE, fill=tkinter.X
        )
        scrollfont.pack(side=tkinter.LEFT, fill=tkinter.Y)
        scorefr.pack()
        wssf = tkinter.Frame(master=fontfr)
        self.s_family = tkinter.Label(
            master=wssf, text=self.tags_variations_comments_font["family"]
        )
        self.s_family.grid_configure(
            column=1, row=1, sticky=tkinter.EW, columnspan=3
        )
        self.s_weight = tkinter.IntVar()
        self._make_font_selector(
            wssf, "Bold:", self.s_weight, 3, self._set_score_font_weight
        )
        self.s_weight.set(self._get_combination("weight", tkinter.font.BOLD))
        self.s_slant = tkinter.IntVar()
        self._make_font_selector(
            wssf, "Italic:", self.s_slant, 6, self._set_score_font_slant
        )
        self.s_slant.set(self._get_combination("slant", tkinter.font.ITALIC))
        wssf.grid_rowconfigure(0, minsize=5)
        wssf.grid_rowconfigure(2, minsize=5)
        wssf.grid_rowconfigure(5, minsize=5)
        wssf.grid_rowconfigure(8, minsize=5)
        wssf.grid_rowconfigure(10, minsize=5)
        wssf.grid_columnconfigure(1, weight=1, uniform="swght")
        wssf.grid_columnconfigure(2, weight=1, uniform="swght")
        wssf.grid_columnconfigure(3, weight=1, uniform="swght")
        self.s_size = tkinter.IntVar()
        tkinter.Label(master=wssf, text="Size:").grid_configure(
            column=0, row=9
        )
        sizeframe = tkinter.Frame(master=wssf)
        fsize = self.tags_variations_comments_font.cget("size")
        sizeframe.grid_configure(
            column=1, row=9, columnspan=3, sticky=tkinter.EW
        )
        for index, item in enumerate(
            (7, 8, 9, 10, 11, 12, 13, 14, 16, 18, 20)
        ):
            sbutton = tkinter.Radiobutton(
                master=sizeframe,
                text=str(item),
                variable=self.s_size,
                value=item,
                indicatoron=tkinter.FALSE,
                command=self.try_command(self._set_score_font_size, sizeframe),
            )
            if item == abs(fsize):
                self.s_size.set(item)
            sbutton.grid_configure(column=index + 1, row=0, sticky=tkinter.EW)
            sizeframe.grid_columnconfigure(index + 1, weight=1, uniform="fsb")
            self.bind(
                sbutton,
                "<ButtonPress>",
                function=self.try_event(focus(sbutton)),
            )
        wssf.pack(fill=tkinter.X)
        fontfr.grid_configure(row=1, column=1)
        self.bind(
            self.s_families,
            "<<ListboxSelect>>",
            function=self.try_event(self._set_score_font_family),
        )

    def _create_game_frame(self):
        """Create sample game for demostrating fonts and colours."""
        self.game = game.Game(
            self.chooser,
            ui=self.ui,
            boardfont=self.boardfont,
            tags_variations_comments_font=self.tags_variations_comments_font,
            moves_played_in_game_font=self.moves_played_in_game_font,
        )

        def focus(event=None):
            del event
            self.game.score.focus_set()

        self.bind(
            self.game.score, "<ButtonPress>", function=self.try_event(focus)
        )
        gamescore = "".join(
            (
                '[Event"National Club: Gosport - Wood Green"]',
                '[Site"Gosport"]',
                '[Date"1989-05-07"]',
                '[Round"QFinal"]',
                '[White"Sowray P J"]',
                '[Black"Marsh R"]',
                '[Result"1-0"]',
                "e4c6d4d5exd5cxd5c4Nf6c5(Nf3)(Nc3)e6Nc3b6b4a5",
                "Bf4axb4Nb5Na6Qa4Bd7",
                "Bc7Nxc5Qd1Qc8dxc5bxc5Nf3Qb7Nd6Bxd6Bxd6Qb6",
                "Be5Ke7Be2Ne4O-Of6Bb2Nc3",
                "{A comment\non two lines}",
                "Bxc3bxc3Qd3(Qb3)Ra3Rfb1Qa7Qc2g6Rb3d4Bc4Rxb3",
                "Bxb3Qa6a4Rb8a5e5Bd5Rb2",
                "Qe4Bf5Qh4Qd3(c2(g5)Nd2Qxa5Rxa5Rb1Nf1Rxf1",
                "Kxf1Bd3)g4Rb1Rxb1Qxb1Kg2",
                "Kd6Qxf6Kxd5Qxe5Kc6gxf5Qxf5Qe8Kc7Qe7Kc8Ne5",
                "c2Qxc5Kd8Qxd4Ke8Qe3Kf8",
                "Kg3Qc8Nd3Kg8f4Qc6Nc1Qa4Qb31-0",
                "\n",
            )
        )
        self.game.collected_game = next(PGN().read_games(gamescore))
        self.game.set_and_tag_item_text()
        self.game.get_top_widget().pack(fill=tkinter.BOTH, expand=1)
        self.game.get_top_widget().pack_propagate(False)

    def get_options(self):
        """Return (olddefaults, newdefaults) colours."""
        cg = game.Game
        cb = board.Board
        olddefaults = {
            constants.LITECOLOR_NAME: cb.litecolor,
            constants.DARKCOLOR_NAME: cb.darkcolor,
            constants.WHITECOLOR_NAME: cb.whitecolor,
            constants.BLACKCOLOR_NAME: cb.blackcolor,
            constants.LINE_COLOR_NAME: cg.l_color,
            constants.MOVE_COLOR_NAME: cg.m_color,
            constants.ALTERNATIVE_MOVE_COLOR_NAME: cg.am_color,
            constants.VARIATION_COLOR_NAME: cg.v_color,
        }
        for fontname in options.font_names:
            olddefaults[fontname] = dict()
        fonts.copy_font_attributes(
            tkinter.font.nametofont(
                constants.MOVES_PLAYED_IN_GAME_FONT
            ).actual(),
            olddefaults[constants.MOVES_PLAYED_IN_GAME_FONT],
        )
        fonts.copy_font_attributes(
            tkinter.font.nametofont(
                constants.TAGS_VARIATIONS_COMMENTS_FONT
            ).actual(),
            olddefaults[constants.TAGS_VARIATIONS_COMMENTS_FONT],
        )
        fonts.copy_board_font_attributes(
            tkinter.font.nametofont(constants.PIECES_ON_BOARD_FONT).actual(),
            olddefaults[constants.PIECES_ON_BOARD_FONT],
        )
        fonts.copy_board_font_attributes(
            tkinter.font.nametofont(
                constants.WILDPIECES_ON_BOARD_FONT
            ).actual(),
            olddefaults[constants.WILDPIECES_ON_BOARD_FONT],
        )
        newdefaults = {
            constants.LITECOLOR_NAME: self.game.board.litecolor,
            constants.DARKCOLOR_NAME: self.game.board.darkcolor,
            constants.WHITECOLOR_NAME: self.game.board.whitecolor,
            constants.BLACKCOLOR_NAME: self.game.board.blackcolor,
            constants.LINE_COLOR_NAME: self.game.l_color,
            constants.MOVE_COLOR_NAME: self.game.m_color,
            constants.ALTERNATIVE_MOVE_COLOR_NAME: self.game.am_color,
            constants.VARIATION_COLOR_NAME: self.game.v_color,
        }
        for fontname in options.font_names:
            newdefaults[fontname] = dict()
        fonts.copy_font_attributes(
            self.tags_variations_comments_font.actual(),
            newdefaults[constants.TAGS_VARIATIONS_COMMENTS_FONT],
        )
        fonts.copy_font_attributes(
            self.moves_played_in_game_font.actual(),
            newdefaults[constants.MOVES_PLAYED_IN_GAME_FONT],
        )
        fonts.copy_board_font_attributes(
            self.boardfont.actual(),
            newdefaults[constants.PIECES_ON_BOARD_FONT],
        )
        fonts.copy_board_font_attributes(
            self.wildpiecesfont.actual(),
            newdefaults[constants.WILDPIECES_ON_BOARD_FONT],
        )
        return olddefaults, newdefaults

    def _get_combination(self, option, value):
        """Return numeric code for move and non-move font combination."""
        tvc = self.tags_variations_comments_font[option] == value
        movesplayed = self.moves_played_in_game_font[option] == value
        if tvc:
            if movesplayed:
                return 2
            return 4
        if movesplayed:
            return 1
        return 3

    def is_ok(self):
        """Return True if dialogue closed using Ok button."""
        return self._ok

    def _on_cancel(self, event=None):
        """Process Cancel button event."""
        del event
        self._ok = False
        self.restore_focus.focus_set()
        self.chooser.destroy()

    def _on_ok(self, event=None):
        """Process Ok button event."""
        del event
        self._ok = True
        self.restore_focus.focus_set()
        self.chooser.destroy()

    def _modal_dialogue(self):
        """Display widget as a modal dialogue.

        Activate game navigation so fonts and colours can be seen in game
        context.

        """
        self.restore_focus = self.chooser.focus_get()
        self.chooser.wait_visibility()
        self.chooser.grab_set()
        self.chooser.wait_window()

    def _set_board_font_family(self, event=None):
        """Set sample game board family."""
        del event
        family = self.b_families
        self.boardfont["family"] = family.get(family.curselection()[0])
        self.game.board.font["family"] = family.get(family.curselection()[0])
        self.b_family["text"] = family.get(family.curselection()[0])

    def _set_board_font_weight(self):
        """Set sample game board pieces font."""
        weights = {0: "normal", 1: "bold"}
        bfweight = weights[self.b_weight.get()]
        self.boardfont["weight"] = bfweight
        self.game.board.font["weight"] = bfweight

    def _set_lite_square_bgcolour(self, colour):
        """Set light squares to colour."""
        self.game.board.litecolor = colour
        self.game.board.set_color_scheme()

    def _set_default_source_for_object(self, colour):
        """Set dark squares to colour."""
        self.game.board.darkcolor = colour
        self.game.board.set_color_scheme()

    def _set_white_piece_fgcolour(self, colour):
        """Set white pieces to colour."""
        self.game.board.whitecolor = colour
        self.game.board.draw_board()

    def _set_black_piece_fgcolour(self, colour):
        """Set black pieces to colour."""
        self.game.board.blackcolor = colour
        self.game.board.draw_board()

    def _set_tag_rest_of_line(self, colour):
        """Set moves following current move in line to colour."""
        self.game.l_color = colour
        self.game.score.tag_configure("l_color", background=colour)

    def _set_tag_move(self, colour):
        """Set current move to colour."""
        self.game.m_color = colour
        self.game.score.tag_configure("m_color", background=colour)

    def _set_tag_alternatives(self, colour):
        """Set alternatives to selected next move to colour."""
        self.game.am_color = colour
        self.game.score.tag_configure("am_color", background=colour)

    def _set_tag_start_of_line(self, colour):
        """Set moves preceeding current move in line to colour."""
        self.game.v_color = colour
        self.game.score.tag_configure("v_color", background=colour)

    def _set_score_font_family(self, event=None):
        """Set game score and variation font family."""
        del event
        family = self.s_families
        self.tags_variations_comments_font["family"] = family.get(
            family.curselection()[0]
        )
        self.moves_played_in_game_font["family"] = family.get(
            family.curselection()[0]
        )
        self.wildpiecesfont["family"] = family.get(family.curselection()[0])
        self.s_family["text"] = family.get(family.curselection()[0])

    def _set_score_font_size(self):
        """Set game score and variation font size to selected size."""
        self.tags_variations_comments_font["size"] = -self.s_size.get()
        self.moves_played_in_game_font["size"] = -self.s_size.get()

    def _set_score_font_slant(self):
        """Set game score and variation font slants to selected slants."""
        slant = {
            4: ("italic", "roman"),
            2: ("italic", "italic"),
            3: ("roman", "roman"),
            1: ("roman", "italic"),
        }
        tvcfslant, mfslant = slant[self.s_slant.get()]
        self.tags_variations_comments_font["slant"] = tvcfslant
        self.moves_played_in_game_font["slant"] = mfslant

    def _set_score_font_weight(self):
        """Set game score and variation font weights to selected weights."""
        weights = {
            4: ("bold", "normal"),
            2: ("bold", "bold"),
            3: ("normal", "normal"),
            1: ("normal", "bold"),
        }
        tvcfweight, mfweight = weights[self.s_weight.get()]
        self.tags_variations_comments_font["weight"] = tvcfweight
        self.moves_played_in_game_font["weight"] = mfweight

    def __del__(self):
        """Force IsOk to return False after __del__ call."""
        self._ok = False
        super().__del__()

    def _make_font_selector(self, frame, title, variable, baserow, command):
        """Create a font selector."""
        tkinter.Label(master=frame, text=title).grid_configure(
            column=0, row=baserow
        )

        def focus(widget):
            def fset(event=None):
                del event
                widget.focus_set()

            return fset

        for index, text in (
            (1, "Moves Played"),
            (2, "All"),
            (3, "None"),
        ):
            radiobutton = tkinter.Radiobutton(
                master=frame,
                text=text,
                variable=variable,
                value=index,
                command=self.try_command(command, frame),
                indicatoron=tkinter.FALSE,
            )
            radiobutton.grid_configure(
                column=index, row=baserow, sticky=tkinter.EW
            )
            self.bind(
                radiobutton,
                "<ButtonPress>",
                function=self.try_event(focus(radiobutton)),
            )
        radiobutton = tkinter.Radiobutton(
            master=frame,
            text="Tags Variations Comments",
            variable=variable,
            value=4,
            command=self.try_command(command, frame),
            indicatoron=tkinter.FALSE,
        )
        radiobutton.grid_configure(
            column=1, row=baserow + 1, columnspan=3, sticky=tkinter.EW
        )
        self.bind(
            radiobutton,
            "<ButtonPress>",
            function=self.try_event(focus(radiobutton)),
        )


class ColourChooser(_ColourScheme):
    """Create a colour selection widget.

    Use this widget on screens up to 950 pixels high.

    """

    def __init__(self, **ka):
        """Extend to create a widget containing just the colour selectors.

        **ka
        ui is the ChessUI instance
        moves tags and pieces are fonts to be used as basis for modified fonts.

        """
        super().__init__(height=660, title="Chessboard colour chooser", **ka)
        self._create_game_frame()
        self._create_colour_frame()
        self._create_buttons(self._get_button_definitions())
        self._modal_dialogue()


class FontChooser(_ColourScheme):
    """Create a font selection widget.

    Use this widget on screens up to 950 pixels high.

    """

    def __init__(self, **ka):
        """Extend to create a widget containing just the font selectors.

        **ka
        ui is the ChessUI instance
        moves tags and pieces are fonts to be used as basis for modified fonts.

        """
        super().__init__(height=680, title="Chessboard font chooser", **ka)
        self._create_font_frame()
        self._create_game_frame()
        self._create_buttons(self._get_button_definitions())
        self._modal_dialogue()


class FontColourChooser(_ColourScheme):
    """Create a font and colour selection widget.

    Use this widget on screens over 950 pixels high.

    """

    def __init__(self, **ka):
        """Extend to create a widget containing font and colour selectors.

        **ka
        ui is the ChessUI instance
        moves tags and pieces are fonts to be used as basis for modified fonts.

        """
        super().__init__(height=950, title="Chessboard style chooser", **ka)
        self._create_font_frame()
        self._create_game_frame()
        self._create_colour_frame()
        self._create_buttons(self._get_button_definitions())
        self._modal_dialogue()


class ChessColourSlider(ColourSlider):
    """Create colour selection widget containing Red Green and Blue sliders."""

    def __init__(
        self,
        master=None,
        row=None,
        label="",
        resolution=2,
        colour="grey",
        tag_onfigure_bachground_colour=None,
    ):
        """Extend with callback to attach a colour to a Text tag."""
        super().__init__(
            master=master,
            row=row,
            label=label,
            resolution=resolution,
            colour=colour,
        )
        self.tag_onfigure_bachground_colour = tag_onfigure_bachground_colour

    def delta_red_colour(self, event=None):
        """Extend to attach colour to Text tag after change to Red."""
        super().delta_red_colour(event=event)
        self.tag_onfigure_bachground_colour(self.get_colour())

    def set_red_colour(self, event=None):
        """Extend to attach colour to Text tag after setting Red."""
        super().set_red_colour(event=event)
        self.tag_onfigure_bachground_colour(self.get_colour())

    def delta_green_colour(self, event=None):
        """Extend to attach colour to Text tag after change to Green."""
        super().delta_green_colour(event=event)
        self.tag_onfigure_bachground_colour(self.get_colour())

    def set_green_colour(self, event=None):
        """Extend to attach colour to Text tag after setting Green."""
        super().set_green_colour(event=event)
        self.tag_onfigure_bachground_colour(self.get_colour())

    def delta_blue_colour(self, event=None):
        """Extend to attach colour to Text tag after change to Blue."""
        super().delta_blue_colour(event=event)
        self.tag_onfigure_bachground_colour(self.get_colour())

    def set_blue_colour(self, event=None):
        """Extend to attach colour to Text tag after setting Blue."""
        super().set_blue_colour(event=event)
        self.tag_onfigure_bachground_colour(self.get_colour())



# chesslmdb.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database using Symas LMMD."""

import os

from solentware_base import lmdb_database
from solentware_base.core import constants

from ..core.filespec import (
    FileSpec,
    GAMES_FILE_DEF,
    LMMD_MINIMUM_FREE_PAGES_AT_START,
)
from ..basecore import database


class ChessDatabase(database.Database, lmdb_database.Database):
    """Provide access to a database of games of chess."""

    _deferred_update_process = os.path.join(
        os.path.basename(os.path.dirname(__file__)), "runchesslmdbdu.py"
    )

    def __init__(
        self,
        DBfile,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        Arguments are passed through to superclass __init__.

        """
        dbnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )

        super().__init__(dbnames, folder=DBfile, **kargs)
        self.open_database()
        (
            map_bytes, map_pages, used_bytes, used_pages, stats
        ) = self.database_stats_summary()
        self.close_database()
        self.map_blocks = map_pages // constants.DEFAULT_MAP_PAGES
        if map_pages - used_pages < LMMD_MINIMUM_FREE_PAGES_AT_START:
            short = (
                LMMD_MINIMUM_FREE_PAGES_AT_START + map_pages - used_pages
            )
        else:
            short = 0
        self.map_blocks = (
            1
            + (map_pages + short) // constants.DEFAULT_MAP_PAGES
        )

    def _delete_database_names(self):
        """Override and return tuple of filenames to delete."""
        return (self.database_file,)
